# SPDX-License-Identifier: AGPL-3.0-or-later
"""
Application manifest for WireGuard.
"""

from django.utils.translation import gettext_lazy as _

from plinth.clients import store_url

_wireguard_package_id = 'com.wireguard.android'

clients = [{
    'name':
        _('WireGuard'),
    'platforms': [{
        'type': 'download',
        'os': 'windows',
        'url': ('https://download.wireguard.com/'
                'windows-client/wireguard-installer.exe')
    }, {
        'type': 'download',
        'os': 'macos',
        'url': 'https://apps.apple.com/us/app/wireguard/id1451685025'
    }, {
        'type': 'package',
        'format': 'deb',
        'name': 'wireguard'
    }, {
        'type': 'package',
        'format': 'rpm',
        'name': 'wireguard-tools'
    }, {
        'type': 'package',
        'format': 'brew',
        'name': 'wireguard-tools'
    }, {
        'type': 'download',
        'os': 'android',
        'url': ('https://www.wireguard.com/'
                'install/#android-play-store-direct-apk-file')
    }, {
        'type': 'store',
        'os': 'android',
        'store_name': 'google-play',
        'url': store_url('google-play', _wireguard_package_id)
    }, {
        'type': 'store',
        'os': 'ios',
        'store_name': 'app-store',
        'url': 'https://apps.apple.com/us/app/wireguard/id1441195209'
    }]
}]

tags = [_('VPN client'), _('VPN server'), _('Privacy'), _('Remote access')]



# SPDX-License-Identifier: AGPL-3.0-or-later
"""
Test module for actions utilities that modify configuration.

Verify that privileged actions perform as expected. See actions.py for a full
description of the expectations.

"""

import os
import typing
from unittest.mock import Mock, call, patch

import pytest

from plinth import actions
from plinth.actions import privileged, secret_str

actions_name = 'actions'


@pytest.fixture(name='popen')
def fixture_popen():
    """A fixture to patch subprocess.Popen called by privileged action."""

    with patch('subprocess.Popen') as popen:

        def call_popen(command, **kwargs):
            write_fd = int(command[8])
            if not isinstance(popen.called_with_write_fd, list):
                popen.called_with_write_fd = []

            popen.called_with_write_fd.append(write_fd)
            os.write(write_fd, bytes(popen.return_value, encoding='utf-8'))
            proc = Mock()
            proc.communicate.return_value = (b'', b'')
            proc.returncode = 0
            return proc

        popen.side_effect = call_popen
        yield popen


def test_privileged_properties():
    """Test that privileged decorator sets proper properties on the method."""

    def func():
        return

    wrapped_func = privileged(func)
    assert wrapped_func._privileged
    assert wrapped_func.__wrapped__ == func


def test_privileged_argument_vararg_check():
    """Test that privileged decorator checks for simple arguments."""

    def func_with_varargs(*_args):
        return

    def func_with_kwargs(**_kwargs):
        return

    def func_with_kwonlyargs(*_args, _kwarg):
        return

    def func_with_kwonlydefaults(*_args, _kwargs='foo'):
        return

    for func in (func_with_varargs, func_with_kwargs, func_with_kwonlyargs,
                 func_with_kwonlydefaults):
        with pytest.raises(SyntaxError):
            privileged(func)


def test_privileged_argument_annotation_check():
    """Test that privileged decorator checks for annotations to arguments."""

    def func1(_a):
        return

    def func2(_a: int, _b):
        return

    # Parameter with 'password' in the name should be typed 'secret_str'
    def func3(_password: str):
        return

    def func1_valid(_a: int, _b: dict[int, str]):
        return

    def func2_valid(_password: secret_str):
        return

    for func in (func1, func2, func3):
        with pytest.raises(SyntaxError):
            privileged(func)

    privileged(func1_valid)
    privileged(func2_valid)


@patch('plinth.actions._read_from_server')
@patch('plinth.actions._request_to_server')
@patch('plinth.actions._get_privileged_action_module_name')
def test_privileged_method_call(get_module_name, request_to_server,
                                read_from_server):
    """Test that privileged method calls the superuser action properly."""

    def func_with_args(_a: int, _b: str, _c: int = 1, _d: str = 'dval',
                       _e: str = 'eval'):
        return

    get_module_name.return_value = 'tests'
    read_from_server.return_value = {"result": "success", "return": "bar"}
    wrapped_func = privileged(func_with_args)
    return_value = wrapped_func(1, 'bval', None, _d='dnewval')
    assert return_value == 'bar'
    assert request_to_server.mock_calls == [
        call({
            'module': 'tests',
            'action': 'func_with_args',
            'args': (1, 'bval', None),
            'kwargs': {
                '_d': 'dnewval'
            }
        }),
        call().close()
    ]


@patch('plinth.actions._read_from_server')
@patch('plinth.actions._request_to_server')
@patch('plinth.actions._get_privileged_action_module_name')
def test_privileged_method_exceptions(get_module_name, request_to_server,
                                      read_from_server):
    """Test that exceptions on privileged methods are return properly."""

    def func_with_exception():
        raise TypeError('type error')

    get_module_name.return_value = 'tests'
    read_from_server.return_value = {
        'result': 'exception',
        'exception': {
            'module': 'builtins',
            'name': 'TypeError',
            'args': ['type error'],
            'traceback': [''],
            'stdout': '',
            'stderr': ''
        }
    }
    wrapped_func = privileged(func_with_exception)
    with pytest.raises(TypeError, match='type error'):
        wrapped_func()


@patch('importlib.import_module')
@patch('plinth.module_loader.get_module_import_path')
@patch('os.getuid')
def test_call_syntax_checks(getuid, get_module_import_path, import_module):
    """Test that calling a method results in proper syntax checks."""
    privileged_call = actions._privileged_call

    # Module name validation
    getuid.return_value = 0
    with pytest.raises(SyntaxError, match='Invalid module name'):
        privileged_call('foo.bar', 'x-action', {})

    # Module import test
    get_module_import_path.return_value = 'plinth.modules.test_module'
    import_module.side_effect = ModuleNotFoundError
    with pytest.raises(SyntaxError, match='Specified module not found'):
        privileged_call('test_module', 'x-action', {})

    import_module.assert_has_calls(
        [call('plinth.modules.test_module.privileged')])

    # Finding action in a module
    module = type('', (), {})
    import_module.side_effect = None
    import_module.return_value = module
    with pytest.raises(SyntaxError, match='Specified action not found'):
        privileged_call('test_module', 'x-action', {})

    # Checking if action is privileged
    def unprivileged_func():
        pass

    setattr(module, 'func', unprivileged_func)
    with pytest.raises(SyntaxError,
                       match='Specified action is not privileged action'):
        privileged_call('test-module', 'func', {})

    # Argument validation
    @actions.privileged
    def func():
        return 'foo'

    setattr(module, 'func', func)
    with pytest.raises(SyntaxError, match='Invalid arguments format'):
        privileged_call('test-module', 'func', {})

    # Successful call
    return_value = privileged_call('test-module', 'func', {
        'args': [],
        'kwargs': {}
    })
    assert return_value == {'result': 'success', 'return': 'foo'}

    # Exception call
    @actions.privileged
    def exception_func():
        raise RuntimeError('foo exception')

    setattr(module, 'func', exception_func)
    return_value = privileged_call('test-module', 'func', {
        'args': [],
        'kwargs': {}
    })
    assert return_value['result'] == 'exception'
    assert return_value['exception']['module'] == 'builtins'
    assert return_value['exception']['name'] == 'RuntimeError'
    assert return_value['exception']['args'] == ('foo exception', )
    assert isinstance(return_value['exception']['traceback'], list)
    for line in return_value['exception']['traceback']:
        assert isinstance(line, str)


def test_assert_valid_arguments():
    """Test that checking valid arguments works."""
    assert_valid = actions._privileged_assert_valid_arguments

    values = [
        None, [], 10, {}, {
            'args': []
        }, {
            'kwargs': {}
        }, {
            'args': {},
            'kwargs': {}
        }, {
            'args': [],
            'kwargs': []
        }
    ]
    for value in values:
        with pytest.raises(SyntaxError, match='Invalid arguments format'):
            assert_valid(lambda: None, value)

    def func(a: int, b: str, c: int = 3, d: str = 'foo'):
        pass

    with pytest.raises(SyntaxError, match='Too many arguments'):
        assert_valid(func, {'args': [1, 2, 3], 'kwargs': {'c': 3, 'd': 4}})

    with pytest.raises(SyntaxError, match='Too many arguments'):
        assert_valid(func, {'args': [1, 2, 3, 4, 5], 'kwargs': {}})

    with pytest.raises(SyntaxError, match='Too many arguments'):
        assert_valid(func, {
            'args': [],
            'kwargs': {
                'a': 1,
                'b': '2',
                'c': 3,
                'd': '4',
                'e': 5
            }
        })

    with pytest.raises(SyntaxError, match='Argument not provided: b'):
        assert_valid(func, {'args': [1], 'kwargs': {}})

    with pytest.raises(SyntaxError, match='Unknown argument: e'):
        assert_valid(func, {'args': [1, '2'], 'kwargs': {'e': 5}})

    with pytest.raises(SyntaxError, match='Duplicate argument: c'):
        assert_valid(func, {'args': [1, '2', 3], 'kwargs': {'c': 4}})

    with pytest.raises(TypeError, match='Expected type str for arg #1'):
        assert_valid(func, {'args': [1, 2], 'kwargs': {}})

    with pytest.raises(TypeError, match='Expected type int for arg c'):
        assert_valid(func, {'args': [1, '2'], 'kwargs': {'c': '3'}})


def test_assert_valid_type():
    """Test that type validation works as expected."""
    assert_valid = actions._privileged_assert_valid_type

    assert_valid(None, None, typing.Any)

    # Invalid values for int, str, float and Optional
    values = [[1, bool], ['foo', int], [1, str], [1, secret_str], [1, float],
              [1, typing.Optional[str]], [1, str | None],
              [1.1, typing.Union[int, str]], [1.1, int | str], [1, list],
              [1, dict], [[1], list[str]], [{
                  'a': 'b'
              }, dict[int, str]], [{
                  1: 2
              }, dict[int, str]]]
    for value in values:
        with pytest.raises(TypeError):
            assert_valid('arg', *value)

    # Valid values
    assert_valid('arg', True, bool)
    assert_valid('arg', 1, int)
    assert_valid('arg', '1', str)
    assert_valid('arg', '1', secret_str)
    assert_valid('arg', 1.1, float)
    assert_valid('arg', None, typing.Optional[int])
    assert_valid('arg', 1, typing.Optional[int])
    assert_valid('arg', None, int | None)
    assert_valid('arg', 1, int | None)
    assert_valid('arg', 1, typing.Union[int, str])
    assert_valid('arg', '1', typing.Union[int, str])
    assert_valid('arg', 1, int | str)
    assert_valid('arg', '1', int | str)
    assert_valid('arg', [], list[int])
    assert_valid('arg', ['foo'], list[str])
    assert_valid('arg', {}, dict[int, str])
    assert_valid('arg', {1: 'foo'}, dict[int, str])



# SPDX-License-Identifier: AGPL-3.0-or-later
"""
Functional, browser based tests for transmission app.
"""

import os

import pytest
from splinter.exceptions import ElementDoesNotExist

from plinth.tests import functional

pytestmark = [pytest.mark.apps, pytest.mark.transmission]


class TestTransmissionApp(functional.BaseAppTests):
    app_name = 'transmission'
    has_service = False
    has_web = True

    def test_upload_torrent(self, session_browser):
        """Test uploading a torrent to Transmission."""
        functional.app_enable(session_browser, 'transmission')
        _remove_all_torrents(session_browser)
        _upload_sample_torrent(session_browser)
        _assert_number_of_torrents(session_browser, 1)

    @pytest.mark.backups
    def test_backup_restore(self, session_browser):
        """Test backup and restore of app data."""
        functional.app_enable(session_browser, 'transmission')
        _remove_all_torrents(session_browser)
        _upload_sample_torrent(session_browser)
        functional.backup_create(session_browser, 'transmission',
                                 'test_transmission')

        _remove_all_torrents(session_browser)
        functional.backup_restore(session_browser, 'transmission',
                                  'test_transmission')

        assert functional.service_is_running(session_browser, 'transmission')
        _assert_number_of_torrents(session_browser, 1)


def _assert_number_of_torrents(session_browser, torrents_number):
    functional.visit(session_browser, '/transmission')
    assert functional.eventually(
        lambda: torrents_number == _get_number_of_torrents(session_browser))


def _remove_all_torrents(browser):
    """Remove all torrents from transmission."""
    functional.visit(browser, '/transmission')
    while True:
        torrents = browser.find_by_css(
            '#torrent_list .torrent, #torrent-list .torrent')
        if not torrents:
            break

        torrents.first.click()
        functional.eventually(
            browser.is_element_not_present_by_css,
            args=['#toolbar-remove.disabled,#toolbar-delete.disabled'])
        browser.find_by_css('#toolbar-remove,#toolbar-delete').first.click()
        functional.eventually(
            browser.is_element_not_present_by_css,
            args=['#dialog-container[style="display: none;"]'])
        try:
            browser.find_by_id('dialog_confirm_button').click()
        except ElementDoesNotExist:
            browser.find_by_css('.dialog-window button').last.click()

        functional.eventually(browser.is_element_present_by_css,
                              args=['#toolbar-remove.disabled'])


def _upload_sample_torrent(browser):
    """Upload a sample torrent into transmission."""
    functional.visit(browser, '/transmission')
    file_path = os.path.join(os.path.dirname(__file__), 'data',
                             'sample.torrent')
    browser.find_by_id('toolbar-open').click()
    # check if older version of Transmission in Debian Bookworm
    transmission_old = browser.is_element_present_by_id('transmission_body')
    if transmission_old:
        functional.eventually(
            browser.is_element_not_present_by_css,
            args=['#upload_container[style="display: none;"]'])
        browser.attach_file('torrent_files[]', [file_path])
        browser.find_by_id('upload_confirm_button').click()
        functional.eventually(browser.is_element_present_by_css,
                              args=['#torrent_list .torrent'])
    else:
        functional.eventually(browser.is_element_present_by_tag,
                              args=['dialog'])
        browser.attach_file('torrent-files[]', [file_path])
        browser.find_by_css('.dialog-window button').last.click()
        functional.eventually(browser.is_element_present_by_css,
                              args=['#torrent-list .torrent'])


def _get_number_of_torrents(browser):
    """Return the number torrents currently in transmission."""
    return len(
        browser.find_by_css('#torrent_list .torrent, #torrent-list .torrent'))



# SPDX-License-Identifier: AGPL-3.0-or-later
"""Configuration helper for BitTorrent web client."""

import pathlib
import shutil
import time

from plinth import action_utils
from plinth.actions import privileged
from plinth.modules.deluge.utils import Config

DELUGE_CONF_DIR = pathlib.Path('/var/lib/deluged/config/')


def _set_configuration(filename, parameter, value):
    """Set the configuration parameter."""
    with action_utils.service_ensure_stopped('deluge-web'):
        with action_utils.service_ensure_stopped('deluged'):
            with Config(DELUGE_CONF_DIR / filename) as config:
                config.content[parameter] = value


def _get_host_id():
    """Get default host id."""
    try:
        with Config(DELUGE_CONF_DIR / 'hostlist.conf') as config:
            return config.content['hosts'][0][0]
    except FileNotFoundError:
        with Config(DELUGE_CONF_DIR / 'hostlist.conf.1.2') as config:
            return config.content['hosts'][0][0]


@privileged
def get_configuration() -> dict[str, str]:
    """Return the current deluged configuration."""
    with Config(DELUGE_CONF_DIR / 'core.conf') as config:
        download_location = config.content['download_location']

    return {'download_location': download_location}


@privileged
def set_configuration(download_location: str):
    """Set the deluged configuration."""
    _set_configuration('core.conf', 'download_location', download_location)


@privileged
def setup():
    """Perform initial setup for deluge."""
    old_service_path = pathlib.Path('/etc/systemd/system/deluge-web.service')
    if old_service_path.exists():
        old_service_path.unlink(missing_ok=True)

    action_utils.service_daemon_reload()
    action_utils.service_try_restart('deluged')
    action_utils.service_try_restart('deluge-web')

    # Wait until core configuration is available so that status of the app can
    # be shown properly.
    _wait_for_configuration()

    # Configure deluge-web to autoconnect to the default deluged daemon.
    host_id = _get_host_id()
    _set_configuration('web.conf', 'default_daemon', host_id)


def _wait_for_configuration():
    """Wait until configuration file has been created."""
    core_file = DELUGE_CONF_DIR / 'core.conf'
    web_file = DELUGE_CONF_DIR / 'web.conf'
    if core_file.exists() and web_file.exists():
        return

    # deluge-web and deluged create files on first run.
    with action_utils.service_ensure_running('deluged'):
        with action_utils.service_ensure_running('deluge-web'):
            for _ in range(60):
                if core_file.exists() and web_file.exists():
                    break

                time.sleep(1)
            else:
                raise Exception('Unable to setup configuration')


@privileged
def uninstall():
    """Remove configuration and service files."""
    # /etc/default/deluged is removed on purge
    shutil.rmtree(DELUGE_CONF_DIR, ignore_errors=True)



# SPDX-License-Identifier: AGPL-3.0-or-later
"""
Manage storage of repository information in KVStore table in database.
"""

import json
from uuid import uuid1

from plinth import kvstore

# kvstore key for repository store
STORAGE_KEY = 'network_storage'
REQUIRED_FIELDS = ['path', 'storage_type', 'added_by_module']


def get_storages(storage_type=None):
    """Get all repositories from store."""
    storages = kvstore.get_default(STORAGE_KEY, {})
    if storages:
        storages = json.loads(storages)

    if storage_type:
        storages = {
            uuid: storage
            for uuid, storage in storages.items()
            if storage['storage_type'] == storage_type
        }

    return storages


def get(uuid):
    """Return a repository with given UUID from store."""
    storages = get_storages()
    return storages[uuid]


def update_or_add(storage):
    """Update an existing or create a new repository in store."""
    for field in REQUIRED_FIELDS:
        if field not in storage:
            raise ValueError('missing storage parameter: %s' % field)

    existing_storages = get_storages()
    if 'uuid' in storage:
        # Replace the existing storage
        existing_storages[storage['uuid']] = storage
    else:
        uuid = str(uuid1())
        storage['uuid'] = uuid
        existing_storages[uuid] = storage

    kvstore.set(STORAGE_KEY, json.dumps(existing_storages))
    return storage['uuid']


def delete(uuid):
    """Remove a repository from store."""
    storages = get_storages()
    del storages[uuid]
    kvstore.set(STORAGE_KEY, json.dumps(storages))



# SPDX-License-Identifier: AGPL-3.0-or-later
"""
Functional, browser based tests for calibre app.
"""

import pathlib
import time

import pytest

from plinth.tests import functional

pytestmark = [pytest.mark.apps, pytest.mark.calibre]


class TestCalibreApp(functional.BaseAppTests):
    app_name = 'calibre'
    has_service = True
    has_web = True

    def test_add_delete_library(self, session_browser):
        """Test adding/deleting a new library."""
        functional.app_enable(session_browser, 'calibre')
        _delete_library(session_browser, 'FunctionalTest', True)

        _add_library(session_browser, 'FunctionalTest')
        assert _is_library_available(session_browser, 'FunctionalTest')

        _delete_library(session_browser, 'FunctionalTest')
        assert not _is_library_available(session_browser, 'FunctionalTest')

    def test_add_delete_book(self, session_browser):
        """Test adding/delete book in the library."""
        functional.app_enable(session_browser, 'calibre')
        _add_library(session_browser, 'FunctionalTest')
        _delete_book(session_browser, 'FunctionalTest', 'sample.txt', True)

        _add_book(session_browser, 'FunctionalTest', 'sample.txt')
        assert _is_book_available(session_browser, 'FunctionalTest',
                                  'sample.txt')

        _delete_book(session_browser, 'FunctionalTest', 'sample.txt')
        assert not _is_book_available(session_browser, 'FunctionalTest',
                                      'sample.txt')

    @pytest.mark.backups
    def test_backup_restore(self, session_browser):
        """Test backing up and restoring."""
        functional.app_enable(session_browser, 'calibre')
        _add_library(session_browser, 'FunctionalTest')
        functional.backup_create(session_browser, 'calibre', 'test_calibre')
        _delete_library(session_browser, 'FunctionalTest')
        functional.backup_restore(session_browser, 'calibre', 'test_calibre')
        assert _is_library_available(session_browser, 'FunctionalTest')


def _add_library(browser, name):
    """Add a new library."""
    if _is_library_available(browser, name):
        return

    browser.links.find_by_href(
        '/freedombox/apps/calibre/library/create/').first.click()
    browser.find_by_id('id_calibre-name').fill(name)
    functional.submit(browser, form_class='form-calibre')


def _delete_library(browser, name, ignore_missing=False):
    """Delete a library."""
    functional.nav_to_module(browser, 'calibre')
    link = browser.links.find_by_href(
        f'/freedombox/apps/calibre/library/{name}/delete/')
    if not link:
        if ignore_missing:
            return

        raise ValueError('Library not found')

    link.first.click()
    functional.submit(browser, form_class='form-delete')


def _is_library_available(browser, name):
    """Return whether a library is present in the list of libraries."""
    functional.nav_to_module(browser, 'calibre')
    link = browser.links.find_by_href(
        f'/freedombox/apps/calibre/library/{name}/delete/')
    return bool(link)


def _visit_library(browser, name):
    """Open the page for the library."""
    functional.visit(browser, '/calibre/')

    # Calibre interface will be available a short time after restarting the
    # service.
    def _service_available():
        unavailable_xpath = '//h1[contains(text(), "Service Unavailable")]'
        available = not browser.find_by_xpath(unavailable_xpath)
        if not available:
            time.sleep(0.5)
            functional.visit(browser, '/calibre/')

        return available

    functional.eventually(_service_available)

    def _library_available():
        """Refresh until the expected library is available."""
        available = browser.find_by_css(
            f'.calibre-push-button[data-lid="{name}"]')
        if not available:
            time.sleep(0.5)
            functional.visit(browser, '/calibre')

        return available

    functional.eventually(_library_available)
    link = browser.find_by_css(f'.calibre-push-button[data-lid="{name}"]')
    if not link:
        raise ValueError('Library not found')

    link.first.click()
    functional.eventually(browser.find_by_css, ['.book-list-cover-grid'])


def _add_book(browser, library_name, book_name):
    """Add a book to the library through Calibre interface."""
    _visit_library(browser, library_name)
    add_button = browser.find_by_css('a[data-button-icon="plus"]')
    add_button.first.click()

    functional.eventually(browser.find_by_xpath,
                          ['//span[contains(text(), "Add books")]'])
    browser.execute_script(
        '''document.querySelector('input[type="file"]').setAttribute(
        'name', 'test-book-upload');''')

    file_path = pathlib.Path(__file__).parent / f'data/{book_name}'
    browser.attach_file('test-book-upload', [str(file_path)])
    functional.eventually(browser.find_by_xpath,
                          ['//span[contains(text(), "Added successfully")]'])


def _delete_book(browser, library_name, book_name, ignore_missing=False):
    """Delete a book from the library through Calibre interface."""
    _visit_library(browser, library_name)
    book_name = book_name.partition('.')[0]
    book = browser.find_by_xpath(f'//a[contains(@title, "{book_name}")]')
    if not book:
        if ignore_missing:
            return

        raise Exception('Book not found')

    book.first.click()
    delete_button = browser.find_by_css('a[data-button-icon="trash"]')
    delete_button.first.click()

    dialog = browser.find_by_id('modal-container').first
    functional.eventually(lambda: dialog.visible)
    ok_button = browser.find_by_xpath('//span[contains(text(), "OK")]')
    ok_button.first.click()


def _is_book_available(browser, library_name, book_name):
    """Return whether a book is present in Calibre interface."""
    _visit_library(browser, library_name)
    book_name = book_name.partition('.')[0]
    book = browser.find_by_xpath(f'//a[contains(@title, "{book_name}")]')
    return bool(book)



# SPDX-License-Identifier: AGPL-3.0-or-later
"""Test privileged method for networks app."""

from unittest.mock import patch

from .. import privileged


@patch('subprocess.run')
def test_get_interfaces(run):
    """Test returning list of network interfaces in sorted order."""
    run.return_value.stdout = '\n'.join([
        'ethernet:ve-fbx-testing',
        'ethernet:enp39s0',
        'ethernet:enp32s1',
        'ethernet:enp4s1',
        'bridge:virbr0',
        'wifi:wlp41s0',
        'loopback:lo',
    ]).encode()

    interfaces = privileged._get_interfaces()
    assert interfaces == {
        'ethernet': ['enp4s1', 'enp32s1', 'enp39s0', 've-fbx-testing'],
        'bridge': ['virbr0'],
        'wifi': ['wlp41s0'],
        'loopback': ['lo']
    }



# SPDX-License-Identifier: AGPL-3.0-or-later

import subprocess

from plinth.errors import PlinthError


class BorgError(PlinthError):
    """Generic borg errors."""


class BorgRepositoryDoesNotExistError(BorgError):
    """Borg access to a repository works but the repository does not exist."""


class SshError(subprocess.CalledProcessError):
    """Error when running an SSH command."""


class SshfsError(PlinthError):
    """Generic sshfs errors."""


class BorgRepositoryExists(BorgError):
    """A repository at target location already exists during initialization."""


class BorgUnencryptedRepository(BorgError):
    """Attempt to provide password on an unencrypted repository."""


class BorgArchiveExists(BorgError):
    """A archive with the given name already exists in the repository."""


class BorgArchiveDoesNotExist(BorgError):
    """Specified archive does not exist in the repository."""


class BorgBusy(BorgError):
    """Borg could not acquire lock being busy with another operation."""


class BorgNoSpace(BorgError):
    """There is not enough space left on the device to perform operation."""



# SPDX-License-Identifier: AGPL-3.0-or-later
"""Views for the SOGo app."""

from django.contrib import messages
from django.utils.translation import gettext_lazy as _

from plinth.privileged import service as service_privileged
from plinth.views import AppView

from . import forms, privileged


class SOGoAppView(AppView):
    """Server configuration page."""
    app_id = 'sogo'
    form_class = forms.DomainForm

    def get_initial(self):
        """Return the initial values to populate in the form."""
        initial = super().get_initial()
        initial['domain'] = privileged.get_domain()
        return initial

    def form_valid(self, form):
        """Update the settings for changed domain values."""
        old_data = form.initial
        new_data = form.cleaned_data
        if old_data['domain'] != new_data['domain']:
            privileged.set_domain(new_data['domain'])
            service_privileged.try_restart('sogo')
            service_privileged.try_restart('memcached')
            messages.success(self.request, _('Configuration updated'))

        return super().form_valid(form)



# SPDX-License-Identifier: AGPL-3.0-or-later
"""
Functional, browser based tests for dynamicdns app.
"""

import pytest

from plinth.tests import functional

pytestmark = [
    pytest.mark.system, pytest.mark.essential, pytest.mark.domain,
    pytest.mark.dynamicdns
]

_configs = {
    'gnudip1': {
        'service_type': 'gnudip',
        'server': 'localhost',
        'domain': 'freedombox.example.com',
        'username': 'tester',
        'password': 'testingtesting',
    },
    'gnudip2': {
        'service_type': 'gnudip',
        'server': '127.0.0.1',
        'domain': 'freedombox2.example.com',
        'username': 'tester2',
        'password': 'testingtesting2',
    },
    'noip.com': {
        'service_type': 'noip.com',
        'update_url': 'https://localhost/update3/',
        'disable_ssl_cert_check': True,
        'use_http_basic_auth': True,
        'domain': 'freedombox3.example.com',
        'username': 'tester3',
        'password': 'testingtesting3',
        'use_ipv6': True,
    },
    'freedns.afraid.org': {
        'service_type': 'freedns.afraid.org',
        'update_url': 'https://localhost/update5/',
        'disable_ssl_cert_check': False,
        'use_http_basic_auth': False,
        'domain': 'freedombox5.example.com',
        'username': '',
        'password': '',
        'use_ipv6': False,
    },
    'other': {
        'service_type': 'other',
        'update_url': 'https://localhost/update6/',
        'disable_ssl_cert_check': False,
        'use_http_basic_auth': False,
        'domain': 'freedombox6.example.com',
        'username': 'tester6',
        'password': 'testingtesting6',
        'use_ipv6': False,
    },
}


@pytest.fixture(scope='module', autouse=True)
def fixture_background(session_browser):
    """Login."""
    functional.login(session_browser)


class TestDynamicDNSApp(functional.BaseAppTests):
    app_name = 'dynamicdns'
    has_service = False
    has_web = False
    can_uninstall = False
    check_diagnostics = False

    @staticmethod
    def test_capitalized_domain_name(session_browser):
        """Test handling of capitalized domain name."""
        config = dict(_configs['gnudip1'], domain='FreedomBox.example.com')
        _configure(session_browser, config)
        _assert_has_config(session_browser,
                           {'domain': 'freedombox.example.com'})

    @staticmethod
    @pytest.mark.parametrize('config_name', list(_configs.keys()))
    def test_various_form_values(session_browser, config_name):
        """Test feeding various values and check that they are saved."""
        _configure(session_browser, _configs[config_name])
        _assert_has_config(session_browser, _configs[config_name])

    @staticmethod
    @pytest.mark.backups
    def test_backup_restore(session_browser):
        """Test backup and restore of configuration."""
        _configure(session_browser, _configs['gnudip1'])
        functional.backup_create(session_browser, 'dynamicdns',
                                 'test_dynamicdns')

        _configure(session_browser, _configs['gnudip2'])
        functional.backup_restore(session_browser, 'dynamicdns',
                                  'test_dynamicdns')

        _assert_has_config(session_browser, _configs['gnudip1'])


def _configure(browser, config):
    functional.nav_to_module(browser, 'dynamicdns')
    current_domains = _get_domains(browser)
    for domain in current_domains:
        if domain.endswith('.example.com'):
            _delete_domain(browser, domain)

    functional.nav_to_module(browser, 'dynamicdns')
    functional.click_link_by_href(browser,
                                  '/freedombox/sys/dynamicdns/domain/add/')
    for key, value in config.items():
        field_id = f'id_domain-{key}'
        if key == 'service_type':
            browser.find_by_id(field_id).select(value)
        elif isinstance(value, bool):
            if value:
                browser.find_by_id(field_id).check()
            else:
                browser.find_by_id(field_id).uncheck()
        else:
            browser.find_by_id(field_id).fill(value)

    functional.submit(browser, form_class='form-domain')


def _assert_has_config(browser, config):
    functional.nav_to_module(browser, 'dynamicdns')
    link = f'/freedombox/sys/dynamicdns/domain/{config["domain"]}/edit/'
    functional.click_link_by_href(browser, link)
    for key, value in config.items():
        if key == 'password':
            continue

        field_id = f'id_domain-{key}'
        if isinstance(value, bool):
            assert browser.find_by_id(field_id).checked == value
        else:
            assert value == browser.find_by_id(field_id).value


def _get_domains(browser):
    """Return the list of configured domains."""
    functional.nav_to_module(browser, 'dynamicdns')
    elements = browser.find_by_css('.domains-status .domain-name a')
    return [element.text.strip() for element in elements]


def _delete_domain(browser, domain):
    """Delete a given domain."""
    functional.nav_to_module(browser, 'dynamicdns')
    link = f'/freedombox/sys/dynamicdns/domain/{domain}/delete/'
    functional.click_link_by_href(browser, link)
    functional.submit(browser, form_class='form-delete')



# SPDX-License-Identifier: AGPL-3.0-or-later
"""Utilities to run operations and show their progress or failures."""

import enum
import logging
import threading
from collections import OrderedDict
from typing import Callable

from plinth.utils import SafeFormatter

from . import app as app_module

logger = logging.getLogger(__name__)


class Operation:
    """Represent an ongoing or finished activity."""

    class State(enum.Enum):
        """Various states of an operation."""

        WAITING = 'waiting'
        RUNNING = 'running'
        COMPLETED = 'completed'

    def __init__(self, op_id: str, app_id: str, name: str, target: Callable,
                 args: list | None = None, kwargs: dict | None = None,
                 show_message: bool = True, show_notification: bool = False,
                 thread_data: dict | None = None,
                 on_complete: Callable | None = None):
        """Initialize to no operation."""
        self.op_id = op_id
        self.app_id = app_id
        self.name = name
        self.show_message = show_message
        self.show_notification = show_notification

        self.target = target
        self.args = args or []
        self.kwargs = kwargs or {}
        self.on_complete = on_complete

        self.state = Operation.State.WAITING
        self.return_value = None
        self._message: str | None = None
        self.exception: Exception | None = None

        # Operation specific data
        self.thread_data: dict = thread_data or {}

        self.thread = threading.Thread(target=self._catch_thread_errors)
        self.start_event = threading.Event()
        setattr(self.thread, '_operation', self)
        self._update_notification()

    def __str__(self):
        """Return a string representation of the operation."""
        return f'Operation: {self.app_id}: {self.name}'

    def _catch_thread_errors(self):
        """Collect exceptions when running in a thread."""
        self._update_notification()
        try:
            self.return_value = self.target(*self.args, **self.kwargs)
        except Exception as exception:
            logger.exception('Error: %s, %s', self, exception)
            self.exception = exception
        finally:
            self.state = Operation.State.COMPLETED
            self._update_notification()
            # Notify
            if self.on_complete:
                self.on_complete(self)

    def run(self):
        """Run a specified operation in a thread."""
        logger.info('%s: running', str(self))
        self.state = Operation.State.RUNNING
        self.thread.start()
        self.start_event.set()

    def join(self):
        """Block the current thread until the operation is completed.

        Raise an exception if the thread encountered an exception.
        """
        self.start_event.wait()
        self.thread.join()
        if self.exception:
            raise self.exception

        return self.return_value

    @staticmethod
    def get_operation() -> 'Operation':
        """Return the operation associated with this thread."""
        thread = threading.current_thread()
        return thread._operation  # type: ignore [attr-defined]

    def on_update(self, message: str | None = None,
                  exception: Exception | None = None):
        """Call from within the thread to update the progress of operation."""
        if message:
            self._message = message

        if exception:
            self.exception = exception

        self._update_notification()

    @property
    def message(self) -> str | None:
        """Return a message about status of the operation."""
        from django.utils.translation import gettext_noop
        if self._message:  # Progress has been set by the operation itself
            return self._message

        if self.exception:  # Operation resulted in a error.
            return gettext_noop('Error: {name}: {exception}')

        if self.state == Operation.State.WAITING:
            return gettext_noop('Waiting to start: {name}')

        if self.state == Operation.State.RUNNING:
            return '{name}'  # No translation needed

        if self.state == Operation.State.COMPLETED:
            return gettext_noop('Finished: {name}')

        return None

    @property
    def translated_message(self) -> str:
        """Return a message about status of operation after translating.

        Must be called from a web request (UI) thread with user language set so
        that localization is done properly.
        """
        from django.utils.translation import gettext
        message = gettext(self.message)
        data = {'name': gettext(self.name), 'exception': str(self.exception)}
        if self.app_id:
            data['app_name'] = app_module.App.get(self.app_id).info.name

        try:
            message = SafeFormatter().vformat(message, [], data)
        except (KeyError, AttributeError) as error:
            logger.warning(
                'Operation missing required key during translation: %s', error)

        return message

    def _update_notification(self) -> None:
        """Show an updated notification if needed."""
        if not self.show_notification:
            return

        from plinth.notification import Notification
        severity = 'info' if not self.exception else 'error'
        app = app_module.App.get(self.app_id)
        data = {
            'app_name': str(app.info.name),
            'app_icon': app.info.icon,
            'app_icon_filename': app.info.icon_filename,
            'state': self.state.value,
            'exception': str(self.exception) if self.exception else None,
            'name': 'translate:' + str(self.name),
        }
        Notification.update_or_create(
            id=self.app_id + '-operation', app_id=self.app_id,
            severity=severity, title=app.info.name, message=self.message,
            body_template='operation-notification.html', data=data,
            group='admin', dismissed=False)


class OperationsManager:
    """Global handler for all operations and their results."""

    def __init__(self) -> None:
        """Initialize the object."""
        self._operations: OrderedDict[str, Operation] = OrderedDict()
        self._current_operation: Operation | None = None

        # Assume that operations manager will be called from various threads
        # including the callback called from the threads it creates. Ensure
        # that properties don't get corrupted due to race conditions when
        # called from different threads by locking all code that updates them.
        # It is re-entrant lock, meaning it can be re-obtained without blocking
        # when done from the same thread which holds the lock.
        self._lock = threading.RLock()

    def new(self, op_id: str, *args, **kwargs) -> Operation:
        """Create a new operation instance and add to global list."""
        with self._lock:
            if (op_id in self._operations and self._operations[op_id].state !=
                    Operation.State.COMPLETED):
                raise KeyError('Operation in progress/scheduled')

            kwargs['on_complete'] = self._on_operation_complete
            operation = Operation(op_id, *args, **kwargs)
            self._operations[op_id] = operation
            logger.info('%s: added', operation)
            self._schedule_next()
            return operation

    def get(self, op_id: str) -> Operation:
        """Return an operation with given operation ID."""
        with self._lock:
            return self._operations[op_id]

    def _on_operation_complete(self, operation: Operation):
        """Trigger next operation. Called from within previous thread."""
        logger.debug('%s: on_complete called', operation)
        with self._lock:
            self._current_operation = None
            if not operation.show_message:
                # No need to keep it lingering for later collection
                del self._operations[operation.op_id]

            self._schedule_next()

    def _schedule_next(self) -> None:
        """Schedule the next available operation."""
        with self._lock:
            if self._current_operation:
                return

            for operation in self._operations.values():
                if operation.state == Operation.State.WAITING:
                    logger.debug('%s: scheduling', operation)
                    self._current_operation = operation
                    operation.run()
                    break

    def filter(self, app_id: str) -> list[Operation]:
        """Return operations matching a pattern."""
        with self._lock:
            return [
                operation for operation in self._operations.values()
                if operation.app_id == app_id
            ]

    def collect_results(self, app_id: str) -> list[Operation]:
        """Return the finished operations for an app."""
        results: list[Operation] = []
        remaining: OrderedDict[str, Operation] = OrderedDict()

        with self._lock:
            for operation in self._operations.values():
                if (operation.app_id == app_id
                        and operation.state == Operation.State.COMPLETED):
                    results.append(operation)
                else:
                    remaining[operation.op_id] = operation

            if results:
                self._operations = remaining

        return results


manager = OperationsManager()



# SPDX-License-Identifier: AGPL-3.0-or-later
"""Functional, browser based tests for Nextcloud app."""

import time
import urllib

import pytest
from selenium.webdriver.common.keys import Keys

from plinth.tests import functional

pytestmark = [pytest.mark.apps, pytest.mark.nextcloud]

PASSWORD = 'kai8wahTei'


class TestNextcloudApp(functional.BaseAppTests):
    app_name = 'nextcloud'
    has_service = True

    def install_and_setup(self, session_browser):
        """Install app and set the domain."""
        super().install_and_setup(session_browser)
        functional.app_enable(session_browser, self.app_name)
        default_url = functional.config['DEFAULT']['url']
        parse_result = urllib.parse.urlparse(default_url)
        override_domain = parse_result.hostname
        override_domain += f':{parse_result.port}' if parse_result.port else ''
        session_browser.find_by_id('id_override_domain').fill(override_domain)
        session_browser.find_by_id('id_admin_password').fill(PASSWORD)
        functional.submit(session_browser, form_class='form-configuration')

    def test_create_folder(self, session_browser, background):
        """Test that creating a folder works."""
        folder_name = 'TestFolder'
        _visit_files_app(session_browser)

        _remove_folder(session_browser, folder_name)
        assert not _is_folder_present(session_browser, folder_name)

        _create_folder(session_browser, folder_name)
        assert _is_folder_present(session_browser, folder_name)

    def test_backup_restore(self, session_browser):
        """Test that backup and restore operations work on the app."""
        folder_name = 'TestBackupFolder'
        _visit_files_app(session_browser)
        if not _is_folder_present(session_browser, folder_name):
            _create_folder(session_browser, folder_name)

        super().test_backup_restore(session_browser)

        _visit_files_app(session_browser)
        assert _is_folder_present(session_browser, folder_name)


def _login(browser):
    """Login to Nextcloud interface."""
    functional.visit(browser, '/nextcloud/')
    if not browser.find_by_css('.login-form'):
        return

    browser.find_by_id('user').fill('nextcloud-admin')
    browser.find_by_id('password').fill(PASSWORD)
    functional.submit(browser, form_class='login-form')


def _visit_files_app(browser):
    """Login to Nextcloud and visit the files app."""
    _login(browser)
    functional.visit(browser, '/nextcloud/apps/files/files')
    # Close the welcome model dialog if it is present
    browser.find_by_tag('html').first.type(Keys.ESCAPE)
    time.sleep(1)  # Allow the model dialog to close.


def _remove_folder(browser, folder_name):
    """Remove a folder from Nextcloud files app."""
    # Select the folder
    xpath = f'//tr[{_class("files-list__row")} and ' \
        f'.//*[{_class("files-list__row-name-")} and ' \
        f'contains(text(),"{folder_name}")]]' \
        f'//span[{_class("checkbox-content-checkbox")}]'
    matches = browser.find_by_xpath(xpath)
    if not matches:
        return

    matches.first.check()

    # Click on '...' more actions button
    xpath = '//button[@aria-label="Actions"]'
    browser.find_by_xpath(xpath).first.click()

    # Click on 'Delete folder' pop down menu item
    xpath = '//button[.//span[contains(text(),"Delete folder")]]'
    browser.find_by_xpath(xpath).first.click()


def _is_folder_present(browser, folder_name):
    """Return whether a folder with given name exists."""
    xpath = f'//tr[{_class("files-list__row")}]' \
        f'//*[{_class("files-list__row-name-")} and ' \
        f'contains(text(),"{folder_name}")]'
    return bool(browser.find_by_xpath(xpath))


def _class(klass):
    """Return xpath snippet for matching class."""
    return f'contains(concat(" ",@class," "), " {klass} ")'


def _create_folder(browser, folder_name):
    """Create a folder in the Nextcloud files app."""
    # Click on the '+ New' button in the header
    xpath = f'//div[{_class("action-item")}]' \
        '//button[.//*[contains(text(),"New")]]'
    browser.find_by_xpath(xpath).first.click()

    # Click on the 'New folder' pop down menu item
    xpath = f'//button[{_class("action-button")} and ' \
        './span[text()="New folder"]]'
    browser.find_by_xpath(xpath).first.click()

    # Get the 'Create new folder' dialog box
    xpath = f'//div[{_class("dialog__modal")} and ' \
        './/h2[text()="Create new folder"]]'
    dialog = browser.find_by_xpath(xpath).first

    # Enter the folder name into text box 'Folder name'
    xpath = f'.//input[{_class("input-field__input")}]'
    dialog.find_by_xpath(xpath).first.fill(folder_name)

    # Press the 'Create' button
    xpath = './/button[.//*[contains(text(),"Create")]]'
    dialog.find_by_xpath(xpath).first.click()



# SPDX-License-Identifier: AGPL-3.0-or-later
"""Forms for configuring Pagekite."""

import copy

from django import forms
from django.contrib import messages
from django.core import validators
from django.utils.translation import gettext as _
from django.utils.translation import gettext_lazy

from . import privileged, utils


class TrimmedCharField(forms.CharField):
    """Trim the contents of a CharField."""

    def clean(self, value):
        """Clean and validate the field value."""
        if value:
            value = value.strip()

        return super().clean(value)


class ConfigurationForm(forms.Form):
    """Configure PageKite credentials and frontend."""

    server_domain = forms.CharField(
        label=gettext_lazy('Server domain'), required=False,
        help_text=gettext_lazy(
            'Select your pagekite server. Set "pagekite.net" to use '
            'the default pagekite.net server.'), widget=forms.TextInput())
    server_port = forms.IntegerField(
        label=gettext_lazy('Server port'), required=False,
        help_text=gettext_lazy('Port of your pagekite server (default: 80)'))
    kite_name = TrimmedCharField(
        label=gettext_lazy('Kite name'),
        help_text=gettext_lazy('Example: mybox.pagekite.me'), validators=[
            validators.RegexValidator(r'^[\w-]{1,63}(\.[\w-]{1,63})*$',
                                      gettext_lazy('Invalid kite name'))
        ])

    kite_secret = TrimmedCharField(
        label=gettext_lazy('Kite secret'), help_text=gettext_lazy(
            'A secret associated with the kite or the default secret '
            'for your account if no secret is set on the kite.'))

    def save(self, request):
        """Save the form on submission after validation."""

        def _filter(data):
            return {
                key: str(value)
                for key, value in data.items() if key in
                ['kite_name', 'kite_secret', 'server_domain', 'server_port']
            }

        if not self.cleaned_data['server_domain']:
            self.cleaned_data['server_domain'] = 'pagekite.net'

        if not self.cleaned_data['server_port']:
            self.cleaned_data['server_port'] = '80'

        old = _filter(self.initial)
        new = _filter(self.cleaned_data)

        # Let's Encrypt certificate paths use lower-case kite name.
        kite_name = new['kite_name'].lower()

        if old != new:
            frontend = f"{new['server_domain']}:{new['server_port']}"
            privileged.set_config(frontend, kite_name, new['kite_secret'])
            messages.success(request, _('Configuration updated'))

            # Update kite name registered with Name Services module.
            utils.update_names_module()


class BaseCustomServiceForm(forms.Form):
    """Basic form functionality to handle a custom service."""

    choices = [('http', 'http'), ('https', 'https'), ('raw', 'raw')]
    protocol = forms.ChoiceField(choices=choices,
                                 label=gettext_lazy('protocol'))
    frontend_port = forms.IntegerField(
        min_value=0, max_value=65535,
        label=gettext_lazy('external (frontend) port'), required=True)
    backend_port = forms.IntegerField(
        min_value=0, max_value=65535,
        label=gettext_lazy('internal (freedombox) port'))
    subdomains = forms.BooleanField(label=gettext_lazy('Enable Subdomains'),
                                    required=False)

    def convert_formdata_to_service(self, formdata):
        """Add information to make a service out of the form data."""
        # convert integers to str (to compare values with DEFAULT_SERVICES)
        for field in ('frontend_port', 'backend_port'):
            formdata[field] = str(formdata[field])

        # set kitename and kitesecret if not already set
        if 'kitename' not in formdata:
            if 'subdomains' in formdata and formdata['subdomains']:
                formdata['kitename'] = "*.%s" % utils.KITE_NAME
            else:
                formdata['kitename'] = utils.KITE_NAME
        if 'secret' not in formdata:
            formdata['secret'] = utils.KITE_SECRET

        # merge protocol and frontend_port back to one entry (protocol)
        if 'frontend_port' in formdata:
            if formdata['frontend_port'] not in formdata['protocol']:
                formdata['protocol'] = "%s/%s" % (formdata['protocol'],
                                                  formdata['frontend_port'])
        if 'backend_host' not in formdata:
            formdata['backend_host'] = utils.BACKEND_HOST

        return formdata


class DeleteCustomServiceForm(BaseCustomServiceForm):
    """Form to remove custom service."""

    def delete(self, request):
        service = self.convert_formdata_to_service(self.cleaned_data)
        privileged.remove_service(service)
        messages.success(request, _('Deleted custom service'))


class AddCustomServiceForm(BaseCustomServiceForm):
    """Adds the save() method and validation to not add predefined services."""

    def matches_predefined_service(self, formdata):
        """Return whether the user input matches a predefined service."""
        service = self.convert_formdata_to_service(formdata)
        match_found = False
        for predefined_service_obj in utils.PREDEFINED_SERVICES.values():
            # manually add the port to compare predefined with custom services
            # that's due to the (sometimes) implicit port in the configuration
            predefined_service = copy.copy(predefined_service_obj['params'])
            if predefined_service['protocol'] == 'http':
                predefined_service['protocol'] = 'http/80'
            elif predefined_service['protocol'] == 'https':
                predefined_service['protocol'] = 'https/443'

            # The formdata service has additional keys, so we can't compare
            # the dicts directly.
            # instead look whether predefined_service is a subset of service
            if all(service[k] == v for k, v in predefined_service.items()):
                match_found = True
                break
        return match_found

    def clean(self):
        cleaned_data = super().clean()
        try:
            is_predefined = self.matches_predefined_service(cleaned_data)
        except KeyError:
            is_predefined = False
        if is_predefined:
            msg = _('This service is already available as a standard service.')
            raise forms.ValidationError(msg)
        return cleaned_data

    def save(self, request):
        service = self.convert_formdata_to_service(self.cleaned_data)
        try:
            privileged.add_service(service)
            messages.success(request, _('Added custom service'))
        except Exception as exception:
            if "already exists" in str(exception):
                messages.error(request, _('This service already exists'))
            else:
                raise



# SPDX-License-Identifier: AGPL-3.0-or-later
"""
Functional, browser based tests for Shadowsocks Server app.
"""

import pytest

from plinth.tests import functional

pytestmark = [pytest.mark.apps, pytest.mark.shadowsocksserver]


class TestShadowsocksServerApp(functional.BaseAppTests):
    app_name = 'shadowsocksserver'
    has_service = True
    has_web = False

    def install_and_setup(self, session_browser):
        """Install the app and run setup."""
        super().install_and_setup(session_browser)
        _configure(session_browser, 'fakepassword')

    @pytest.mark.backups
    def test_backup_restore(self, session_browser):
        """Test backup and restore of configuration."""
        _configure(session_browser, 'beforebackup123')
        functional.backup_create(session_browser, 'shadowsocksserver',
                                 'test_shadowsocksserver')

        _configure(session_browser, 'afterbackup123')
        functional.backup_restore(session_browser, 'shadowsocksserver',
                                  'test_shadowsocksserver')

        assert functional.service_is_running(session_browser,
                                             'shadowsocksserver')
        assert _get_configuration(session_browser) == 'beforebackup123'


def _configure(browser, password):
    """Configure Shadowsocks Server with given details."""
    functional.visit(browser, '/freedombox/apps/shadowsocksserver/')
    browser.find_by_id('id_password').fill(password)
    functional.submit(browser, form_class='form-configuration')


def _get_configuration(browser):
    """Return the password currently configured in Shadowsocks Server."""
    functional.visit(browser, '/freedombox/apps/shadowsocksserver/')
    password = browser.find_by_id('id_password').value
    return password



# SPDX-License-Identifier: AGPL-3.0-or-later
"""Functional, browser based tests for names app."""

import pytest

from plinth.tests import functional

pytestmark = [
    pytest.mark.system, pytest.mark.essential, pytest.mark.domain,
    pytest.mark.names
]


@pytest.fixture(scope='module', autouse=True)
def fixture_background(session_browser):
    """Login."""
    functional.login(session_browser)


def test_change_hostname(session_browser):
    """Test changing the hostname."""
    functional.set_hostname(session_browser, 'mybox')
    assert _get_hostname(session_browser) == 'mybox'


def _get_hostname(browser):
    functional.visit(browser, '/freedombox/sys/names/hostname/')
    return browser.find_by_id('id_hostname-hostname').value


def test_change_domain_name(session_browser):
    """Test changing the domain name."""
    functional.domain_remove(session_browser, 'mydomain.example')
    functional.domain_add(session_browser, 'mydomain.example')
    assert 'mydomain.example' in functional.domain_list(session_browser)

    # Capitalization is ignored.
    functional.domain_remove(session_browser, 'mydomain2.example')
    functional.domain_add(session_browser, 'Mydomain2.example')
    assert 'mydomain2.example' in functional.domain_list(session_browser)



# SPDX-License-Identifier: AGPL-3.0-or-later
"""Application manifest for miniflux."""

from django.utils.translation import gettext_lazy as _

from plinth.clients import store_url

clients = [
    {
        'name': _('Miniflux'),
        'platforms': [{
            'type': 'web',
            'url': '/miniflux/'
        }]
    },
    {
        'name':
            _('Fluent Reader Lite'),
        'platforms': [{
            'type': 'store',
            'os': 'android',
            'store_name': 'google-play',
            'url': store_url('google-play', 'me.hyliu.fluent_reader_lite'),
        }, {
            'type': 'store',
            'os': 'ios',
            'store_name': 'app-store',
            'url': 'https://apps.apple.com/app/id1549611796',
        }]
    },
    {
        'name':
            _('Fluent Reader'),
        'platforms': [{
            'type': 'download',
            'os': 'windows',
            'url': 'https://github.com/yang991178/fluent-reader/releases',
        }, {
            'type': 'download',
            'os': 'macos',
            'url': 'https://github.com/yang991178/fluent-reader/releases',
        }]
    },
    {
        'name':
            _('FluxNews'),
        'platforms': [{
            'type': 'store',
            'os': 'android',
            'store_name': 'f-droid',
            'url': store_url('f-droid', 'de.circle_dev.flux_news')
        }, {
            'type': 'store',
            'os': 'android',
            'store_name': 'google-play',
            'url': store_url('google-play', 'de.circle_dev.flux_news')
        }]
    },
    {
        'name':
            _('MiniFlutt'),
        'platforms': [{
            'type': 'store',
            'os': 'android',
            'store_name': 'google-play',
            'url': store_url('google-play', 'be.martinelli.miniflutt')
        }]
    },
    {
        'name':
            _('NetNewsWire'),
        'platforms': [{
            'type': 'download',
            'os': 'macos',
            'url': 'https://netnewswire.com/',
        }, {
            'type': 'store',
            'os': 'ios',
            'store_name': 'app-store',
            'url': 'https://apps.apple.com/us/app/'
                   'netnewswire-rss-reader/id1480640210',
        }]
    },
    {
        'name':
            _('Newsflash'),
        'platforms': [{
            'type': 'download',
            'os': 'gnu-linux',
            'url': 'https://flathub.org/apps/details/'
                   'io.gitlab.news_flash.NewsFlash'
        }]
    },
    {
        'name':
            _('Read You'),
        'platforms': [{
            'type': 'store',
            'os': 'android',
            'store_name': 'f-droid',
            'url': store_url('f-droid', 'me.ash.reader')
        }]
    },
    {
        'name':
            _('RSS Guard'),
        'platforms': [{
            'type': 'download',
            'os': 'gnu-linux',
            'url': 'https://github.com/martinrotter/rssguard/releases',
        }, {
            'type': 'download',
            'os': 'macos',
            'url': 'https://github.com/martinrotter/rssguard/releases',
        }, {
            'type': 'download',
            'os': 'windows',
            'url': 'https://github.com/martinrotter/rssguard/releases',
        }]
    },
]

backup = {
    'config': {
        'files': [
            '/etc/miniflux/freedombox.conf',
            '/var/lib/plinth/backups-data/miniflux-database.sql',
        ],
    },
    'secrets': {
        'files': [
            '/etc/miniflux/database', '/etc/dbconfig-common/miniflux.conf'
        ]
    },
    'services': ['miniflux']
}

tags = [_('Feed reader'), _('News aggregation'), _('RSS'), _('ATOM')]



# SPDX-License-Identifier: AGPL-3.0-or-later
"""FreedomBox app to interface with network-manager."""

import subprocess
from logging import Logger

from django.utils.translation import gettext_lazy as _

from plinth import app as app_module
from plinth import daemon, kvstore, menu, network
from plinth.config import DropinConfigs
from plinth.diagnostic_check import DiagnosticCheck
from plinth.modules.firewall.components import Firewall
from plinth.package import Packages

from . import manifest, privileged

_description = [
    _('Configure network devices. Connect to the Internet via Ethernet, Wi-Fi '
      'or PPPoE. Share that connection with other devices on the network.'),
    _('Devices administered through other methods may not be available for '
      'configuration here.'),
]

logger = Logger(__name__)


class NetworksApp(app_module.App):
    """FreedomBox app for Networks."""

    app_id = 'networks'

    _version = 2

    can_be_disabled = False

    def __init__(self) -> None:
        """Create components for the app."""
        super().__init__()

        info = app_module.Info(app_id=self.app_id, version=self._version,
                               is_essential=True, name=_('Networks'),
                               icon='fa-signal', description=_description,
                               manual_page='Networks', tags=manifest.tags)
        self.add(info)

        menu_item = menu.Menu('menu-networks', info.name, info.icon, info.tags,
                              'networks:index',
                              parent_url_name='system:system', order=20)
        self.add(menu_item)

        packages = Packages('packages-networks', ['network-manager', 'batctl'])
        self.add(packages)

        # For 'shared' network connections
        firewall = Firewall('firewall-networks', info.name,
                            ports=['dns', 'dhcp'], is_external=False)
        self.add(firewall)

        dropin_configs = DropinConfigs('dropin-configs-networks', [
            '/etc/NetworkManager/dispatcher.d/10-freedombox-batman',
        ])
        self.add(dropin_configs)

    def diagnose(self) -> list[DiagnosticCheck]:
        """Run diagnostics and return the results."""
        results = super().diagnose()

        interfaces = _get_shared_interfaces()
        addresses = _get_interface_addresses(interfaces)

        for address in addresses:
            results.append(daemon.diagnose_port_listening(53, 'tcp', address))
            results.append(daemon.diagnose_port_listening(53, 'udp', address))

        return results

    def setup(self, old_version):
        """Install and configure the app."""
        super().setup(old_version)
        privileged.setup()
        self.enable()


def get_network_topology_type():
    """Return the currently configured network topology type or default."""
    return kvstore.get_default('networks_topology_type', 'to_router')


def set_network_topology_type(network_topology_type):
    """Store the network topology type."""
    kvstore.set('networks_topology_type', network_topology_type)


def get_internet_connection_type():
    """Return the currently configured internet connection type or default."""
    return kvstore.get_default('networks_internet_type', 'unknown')


def set_internet_connection_type(internet_connection_type):
    """Store the internet connection type."""
    return kvstore.set('networks_internet_type', internet_connection_type)


def get_router_configuration_type():
    """Return the currently configured router configuration type or default."""
    return kvstore.get_default('networks_router_configuration_type',
                               'not_configured')


def set_router_configuration_type(router_configuration_type):
    """Store the router configuration type."""
    return kvstore.set('networks_router_configuration_type',
                       router_configuration_type)


def _get_shared_interfaces():
    """Get active network interfaces in shared mode."""
    shared_interfaces = []
    for connection in network.get_connection_list():
        if not connection['is_active']:
            continue

        connection_uuid = connection['uuid']
        connection = network.get_connection(connection_uuid)

        settings_ipv4 = connection.get_setting_ip4_config()
        if settings_ipv4.get_method() == 'shared':
            settings_connection = connection.get_setting_connection()
            interface = settings_connection.get_interface_name()
            if interface:
                shared_interfaces.append(interface)

    return shared_interfaces


def _get_interface_addresses(interfaces):
    """Get the IPv4 addresses for the given interfaces."""
    output = subprocess.check_output(['ip', '-o', 'addr'])
    lines = output.decode().splitlines()

    addresses = []
    for line in lines:
        parts = line.split()
        if parts[1] in interfaces and parts[2] == 'inet':
            addresses.append(parts[3].split('/')[0])

    return addresses



# SPDX-License-Identifier: AGPL-3.0-or-later
"""
Test module for network configuration utilities.
"""

import copy
import threading
import time

import pytest

from plinth.utils import import_from_gi

ethernet_settings = {
    'common': {
        'type': '802-3-ethernet',
        'name': 'plinth_test_eth',
        'interface': 'eth0',
        'zone': 'internal',
        'dns_over_tls': 'opportunistic',
    },
    'ipv4': {
        'method': 'auto',
        'dns': '',
        'second_dns': '',
    },
    'ipv6': {
        'method': 'auto',
        'dns': '',
        'second_dns': '',
    },
}

wifi_settings = {
    'common': {
        'type': '802-11-wireless',
        'name': 'plinth_test_wifi',
        'interface': 'wlan0',
        'zone': 'external',
        'dns_over_tls': 'yes',
    },
    'ipv4': {
        'method': 'auto',
        'dns': '',
        'second_dns': '',
    },
    'ipv6': {
        'method': 'auto',
        'dns': '',
        'second_dns': '',
    },
    'wireless': {
        'ssid': 'plinthtestwifi',
        'mode': 'adhoc',
        'auth_mode': 'open',
        'band': 'a',
        'channel': 0,
        'bssid': 'a0:86:c6:08:11:02',
    },
}

pppoe_settings = {
    'common': {
        'type': 'pppoe',
        'name': 'plinth_test_pppoe',
        'interface': 'eth1',
        'zone': 'internal',
    },
    'pppoe': {
        'username': 'x-user',
        'password': 'x-password',
    },
}


@pytest.fixture(autouse=True, scope='module')
def fixture_network_module_init():
    """Initialize network module in a separate thread."""
    from plinth import network as network_module
    glib = import_from_gi('GLib', '2.0')
    main_loop = glib.MainLoop()

    def main_loop_runner():
        """Initialize the network module and run glib main loop until quit."""
        network_module.init()
        main_loop.run()

    thread = threading.Thread(target=main_loop_runner)
    thread.start()

    while not network_module._client:
        time.sleep(0.1)

    yield

    if main_loop:
        main_loop.quit()

    thread.join()


@pytest.fixture(name='network')
def fixture_network(needs_root):
    """Return the network module. Load it conservatively."""
    from plinth import network as network_module
    return network_module


def _connection(network, settings):
    """Create, yield and delete a connection."""
    uuid = network.add_connection(settings)
    time.sleep(0.1)
    yield uuid
    time.sleep(0.1)
    network.delete_connection(uuid)


@pytest.fixture(name='ethernet_uuid')
def fixture_ethernet_uuid(network):
    """Test fixture to setup and teardown an Ethernet connection."""
    yield from _connection(network, ethernet_settings)


@pytest.fixture(name='wifi_uuid')
def fixture_wifi_uuid(network):
    """Test fixture to setup and teardown an Wi-Fi connection."""
    yield from _connection(network, wifi_settings)


@pytest.fixture(name='pppoe_uuid')
def fixture_pppoe_uuid(network):
    """Test fixture to setup and teardown an PPPoE connection."""
    yield from _connection(network, pppoe_settings)


@pytest.mark.usefixtures('ethernet_uuid', 'wifi_uuid', 'pppoe_uuid')
def test_get_connection_list(network):
    """Check that we can get a list of available connections."""
    connections = network.get_connection_list()
    connection_names = [conn['name'] for conn in connections]

    assert 'plinth_test_eth' in connection_names
    assert 'plinth_test_wifi' in connection_names
    assert 'plinth_test_pppoe' in connection_names


def test_get_connection(network, ethernet_uuid, wifi_uuid):
    """Check that we can get a connection by name."""
    connection = network.get_connection(ethernet_uuid)
    assert connection.get_id() == 'plinth_test_eth'

    connection = network.get_connection(wifi_uuid)
    assert connection.get_id() == 'plinth_test_wifi'

    with pytest.raises(network.ConnectionNotFound):
        network.get_connection('x-invalid-network-id')


def test_edit_ethernet_connection(network, ethernet_uuid):
    """Check that we can update an ethernet connection."""
    connection = network.get_connection(ethernet_uuid)
    ethernet_settings2 = copy.deepcopy(ethernet_settings)
    ethernet_settings2['common']['name'] = 'plinth_test_eth_new'
    ethernet_settings2['common']['interface'] = 'eth1'
    ethernet_settings2['common']['zone'] = 'external'
    ethernet_settings2['common']['dns_over_tls'] = 'no'
    ethernet_settings2['common']['autoconnect'] = False
    ethernet_settings2['ipv4']['method'] = 'auto'
    network.edit_connection(connection, ethernet_settings2)

    connection = network.get_connection(ethernet_uuid)
    assert connection.get_id() == 'plinth_test_eth_new'

    settings_connection = connection.get_setting_connection()
    assert settings_connection.get_interface_name() == 'eth1'
    assert settings_connection.get_zone() == 'external'
    assert settings_connection.get_dns_over_tls().value_nick == 'no'
    assert not settings_connection.get_autoconnect()

    settings_ipv4 = connection.get_setting_ip4_config()
    assert settings_ipv4.get_method() == 'auto'


def test_edit_pppoe_connection(network, pppoe_uuid):
    """Check that we can update a PPPoE connection."""
    connection = network.get_connection(pppoe_uuid)
    pppoe_settings2 = copy.deepcopy(pppoe_settings)
    pppoe_settings2['common']['name'] = 'plinth_test_pppoe_new'
    pppoe_settings2['common']['interface'] = 'eth2'
    pppoe_settings2['common']['zone'] = 'external'
    pppoe_settings2['common']['autoconnect'] = False
    pppoe_settings2['pppoe']['username'] = 'x-user-new'
    pppoe_settings2['pppoe']['password'] = 'x-password-new'
    network.edit_connection(connection, pppoe_settings2)

    connection = network.get_connection(pppoe_uuid)
    assert connection.get_id() == 'plinth_test_pppoe_new'

    settings_connection = connection.get_setting_connection()
    assert settings_connection.get_interface_name() == 'eth2'
    assert settings_connection.get_zone() == 'external'
    assert not settings_connection.get_autoconnect()

    settings_pppoe = connection.get_setting_pppoe()
    assert settings_pppoe.get_username() == 'x-user-new'
    secrets = connection.get_secrets('pppoe')
    assert secrets['pppoe']['password'] == 'x-password-new'

    settings_ppp = connection.get_setting_ppp()
    assert settings_ppp.get_lcp_echo_failure() == 5
    assert settings_ppp.get_lcp_echo_interval() == 30


def test_edit_wifi_connection(network, wifi_uuid):
    """Check that we can update a wifi connection."""
    connection = network.get_connection(wifi_uuid)
    wifi_settings2 = copy.deepcopy(wifi_settings)
    wifi_settings2['common']['name'] = 'plinth_test_wifi_new'
    wifi_settings2['common']['interface'] = 'wlan1'
    wifi_settings2['common']['zone'] = 'external'
    wifi_settings2['common']['dns_over_tls'] = 'opportunistic'
    wifi_settings2['common']['autoconnect'] = False
    wifi_settings2['ipv4']['method'] = 'auto'
    wifi_settings2['wireless']['ssid'] = 'plinthtestwifi2'
    wifi_settings2['wireless']['mode'] = 'infrastructure'
    wifi_settings2['wireless']['auth_mode'] = 'wpa'
    wifi_settings2['wireless']['passphrase'] = 'secretpassword'
    network.edit_connection(connection, wifi_settings2)

    connection = network.get_connection(wifi_uuid)

    assert connection.get_id() == 'plinth_test_wifi_new'

    settings_connection = connection.get_setting_connection()
    assert settings_connection.get_interface_name() == 'wlan1'
    assert settings_connection.get_zone() == 'external'
    assert settings_connection.get_dns_over_tls().value_nick == 'opportunistic'
    assert not settings_connection.get_autoconnect()

    settings_wireless = connection.get_setting_wireless()
    assert settings_wireless.get_ssid().get_data() == b'plinthtestwifi2'
    assert settings_wireless.get_ssid().get_data().decode(
    ) == 'plinthtestwifi2'
    assert settings_wireless.get_mode() == 'infrastructure'

    wifi_sec = connection.get_setting_wireless_security()
    assert wifi_sec.get_key_mgmt() == 'wpa-psk'

    secrets = connection.get_secrets('802-11-wireless-security')
    assert secrets['802-11-wireless-security']['psk'] == 'secretpassword'


def test_ethernet_manual_ipv4_address(network, ethernet_uuid):
    """Check that we can manually set IPv4 address on ethernet."""
    connection = network.get_connection(ethernet_uuid)
    ethernet_settings2 = copy.deepcopy(ethernet_settings)
    ethernet_settings2['ipv4']['method'] = 'manual'
    ethernet_settings2['ipv4']['address'] = '169.254.0.1'
    ethernet_settings2['ipv4']['netmask'] = '255.255.254.0'
    ethernet_settings2['ipv4']['gateway'] = '169.254.0.254'
    ethernet_settings2['ipv4']['dns'] = '1.2.3.4'
    ethernet_settings2['ipv4']['second_dns'] = '1.2.3.5'
    network.edit_connection(connection, ethernet_settings2)

    connection = network.get_connection(ethernet_uuid)
    settings_ipv4 = connection.get_setting_ip4_config()
    assert settings_ipv4.get_method() == 'manual'

    address = settings_ipv4.get_address(0)
    assert address.get_address() == '169.254.0.1'
    assert address.get_prefix() == 23
    assert settings_ipv4.get_gateway() == '169.254.0.254'
    assert settings_ipv4.get_num_dns() == 2
    assert settings_ipv4.get_dns(0) == '1.2.3.4'
    assert settings_ipv4.get_dns(1) == '1.2.3.5'


def test_ethernet_manual_ipv6_address(network, ethernet_uuid):
    """Check that we can manually set IPv6 address on ethernet."""
    connection = network.get_connection(ethernet_uuid)
    ethernet_settings2 = copy.deepcopy(ethernet_settings)
    ethernet_settings2['ipv6']['method'] = 'manual'
    ethernet_settings2['ipv6']['address'] = '::ffff:169.254.0.1'
    ethernet_settings2['ipv6']['prefix'] = '63'
    ethernet_settings2['ipv6']['gateway'] = '::ffff:169.254.0.254'
    ethernet_settings2['ipv6']['dns'] = '::ffff:1.2.3.4'
    ethernet_settings2['ipv6']['second_dns'] = '::ffff:1.2.3.5'
    network.edit_connection(connection, ethernet_settings2)

    connection = network.get_connection(ethernet_uuid)
    settings_ipv6 = connection.get_setting_ip6_config()
    assert settings_ipv6.get_method() == 'manual'

    address = settings_ipv6.get_address(0)
    assert address.get_address() == '::ffff:169.254.0.1'
    assert address.get_prefix() == 63
    assert settings_ipv6.get_gateway() == '::ffff:169.254.0.254'
    assert settings_ipv6.get_num_dns() == 2
    assert settings_ipv6.get_dns(0) == '::ffff:1.2.3.4'
    assert settings_ipv6.get_dns(1) == '::ffff:1.2.3.5'


def test_wifi_manual_ipv4_address(network, wifi_uuid):
    """Check that we can manually set IPv4 address on wifi."""
    connection = network.get_connection(wifi_uuid)
    wifi_settings2 = copy.deepcopy(wifi_settings)
    wifi_settings2['ipv4']['method'] = 'manual'
    wifi_settings2['ipv4']['address'] = '169.254.0.2'
    wifi_settings2['ipv4']['netmask'] = '255.255.254.0'
    wifi_settings2['ipv4']['gateway'] = '169.254.0.254'
    wifi_settings2['ipv4']['dns'] = '1.2.3.4'
    wifi_settings2['ipv4']['second_dns'] = '1.2.3.5'
    wifi_settings2['wireless']['ssid'] = 'plinthtestwifi'
    wifi_settings2['wireless']['mode'] = 'adhoc'
    wifi_settings2['wireless']['auth_mode'] = 'open'
    network.edit_connection(connection, wifi_settings2)

    connection = network.get_connection(wifi_uuid)
    settings_ipv4 = connection.get_setting_ip4_config()
    assert settings_ipv4.get_method() == 'manual'

    address = settings_ipv4.get_address(0)
    assert address.get_address() == '169.254.0.2'
    assert address.get_prefix() == 23
    assert settings_ipv4.get_gateway() == '169.254.0.254'
    assert settings_ipv4.get_num_dns() == 2
    assert settings_ipv4.get_dns(0) == '1.2.3.4'
    assert settings_ipv4.get_dns(1) == '1.2.3.5'


def test_wifi_manual_ipv6_address(network, wifi_uuid):
    """Check that we can manually set IPv6 address on wifi."""
    connection = network.get_connection(wifi_uuid)
    wifi_settings2 = copy.deepcopy(wifi_settings)
    wifi_settings2['ipv6']['method'] = 'manual'
    wifi_settings2['ipv6']['address'] = '::ffff:169.254.0.2'
    wifi_settings2['ipv6']['prefix'] = 63
    wifi_settings2['ipv6']['gateway'] = '::ffff:169.254.0.254'
    wifi_settings2['ipv6']['dns'] = '::ffff:1.2.3.4'
    wifi_settings2['ipv6']['second_dns'] = '::ffff:1.2.3.5'
    wifi_settings2['wireless']['ssid'] = 'plinthtestwifi'
    wifi_settings2['wireless']['mode'] = 'adhoc'
    wifi_settings2['wireless']['auth_mode'] = 'open'
    network.edit_connection(connection, wifi_settings2)

    connection = network.get_connection(wifi_uuid)
    settings_ipv6 = connection.get_setting_ip6_config()
    assert settings_ipv6.get_method() == 'manual'

    address = settings_ipv6.get_address(0)
    assert address.get_address() == '::ffff:169.254.0.2'
    assert address.get_prefix() == 63
    assert settings_ipv6.get_gateway() == '::ffff:169.254.0.254'
    assert settings_ipv6.get_num_dns() == 2
    assert settings_ipv6.get_dns(0) == '::ffff:1.2.3.4'
    assert settings_ipv6.get_dns(1) == '::ffff:1.2.3.5'



# SPDX-License-Identifier: AGPL-3.0-or-later
"""
Utility methods for providing client information.
"""

from enum import Enum

from django.utils.functional import Promise


class Desktop_OS(Enum):
    GNU_LINUX = 'gnu-linux'
    MAC_OS = 'macos'
    WINDOWS = 'windows'


class Mobile_OS(Enum):
    ANDROID = 'android'
    IOS = 'ios'


class Store(Enum):
    APP_STORE = 'app-store'
    F_DROID = 'f-droid'
    GOOGLE_PLAY = 'google-play'


class Package(Enum):
    DEB = 'deb'
    HOMEBREW = 'brew'
    RPM = 'rpm'


def enum_values(enum):
    return [option.value for option in list(enum)]


def _check(client, condition):
    """Check if any of a list of clients satisfies the given condition"""
    return any(platform for platform in client['platforms']
               if condition(platform))


def _client_has_desktop(client):
    """Filter to find out whether an application has desktop clients"""
    return _check(
        client, lambda platform: platform.get('os') in enum_values(Desktop_OS)
        and platform.get('type') != 'package')


def _client_has_mobile(client):
    """Filter to find out whether an application has mobile clients"""
    return _check(
        client, lambda platform: platform.get('os') in enum_values(Mobile_OS))


def _client_has_web(client):
    """Filter to find out whether an application has web clients"""
    return _check(client, lambda platform: platform['type'] == 'web')


def _client_has_package(client):
    """Filter to find out whether an application has web clients"""
    return _check(client, lambda platform: platform['type'] == 'package')


def of_type(clients, client_type):
    """Filter and get clients of a particular type"""
    filters = {
        'mobile': _client_has_mobile,
        'desktop': _client_has_desktop,
        'web': _client_has_web,
        'package': _client_has_package,
    }
    return list(filter(filters[client_type], clients))


def store_url(store, package_id):
    """Return a full App store URL given package id and type of store."""
    stores = {
        'google-play': 'https://play.google.com/store/apps/details?id={}',
        'f-droid': 'https://f-droid.org/packages/{}'
    }
    return stores[store].format(package_id)


def validate(clients):
    """Validate the clients' information schema."""
    assert isinstance(clients, list)
    for client in clients:
        _validate_client(client)

    return clients


def _validate_client(client):
    """Validate a single client's information schema."""
    assert isinstance(client, dict)
    assert 'name' in client
    assert isinstance(client['platforms'], list)
    for platform in client['platforms']:
        _validate_platform(platform)


def _validate_platform(platform):
    """Validate a single platform's schema."""
    assert platform['type'] in ('package', 'download', 'store', 'web')
    validate_method = globals()['_validate_platform_' + platform['type']]
    validate_method(platform)


def _validate_platform_package(platform):
    """Validate a platform of type package."""
    assert platform['format'] in enum_values(Package)
    assert isinstance(platform['name'], (str, Promise))


def _validate_platform_download(platform):
    """Validate a platform of type download."""
    assert platform['os'] in enum_values(Desktop_OS) + enum_values(Mobile_OS)
    assert isinstance(platform['url'], (str, Promise))


def _validate_platform_store(platform):
    """Validate a platform of type store."""
    assert platform['os'] in enum_values(Mobile_OS)
    assert platform['store_name'] in enum_values(Store)
    assert isinstance(platform['url'], (str, Promise))


def _validate_platform_web(platform):
    """Validate a platform of type web."""
    assert isinstance(platform['url'], (str, Promise))



# SPDX-License-Identifier: AGPL-3.0-or-later
"""
Functional, browser based tests for gitweb app.
"""
import contextlib
import os
import shutil
import subprocess
import tempfile

import pytest

from plinth.tests import functional

pytestmark = [pytest.mark.apps, pytest.mark.gitweb]

_default_url = functional.config['DEFAULT']['url']


class TestGitwebApp(functional.BaseAppTests):
    app_name = 'gitweb'
    has_service = False
    has_web = True

    def test_all_repos_private(self, session_browser):
        """Test repo accessability when all repos are private."""
        _create_repo(session_browser, 'Test-repo', 'private',
                     ok_if_exists=True)
        _set_all_repos_private(session_browser)
        if not functional.user_exists(session_browser, 'gitweb_user'):
            functional.create_user(session_browser, 'gitweb_user',
                                   groups=['git-access'])
        if not functional.user_exists(session_browser, 'nogroupuser'):
            functional.create_user(session_browser, 'nogroupuser', groups=[])

        functional.login_with_account(session_browser, functional.base_url,
                                      'gitweb_user')
        assert functional.is_available(session_browser, 'gitweb')
        assert len(functional.find_on_front_page(session_browser,
                                                 'gitweb')) == 1

        functional.login_with_account(session_browser, functional.base_url,
                                      'nogroupuser')
        assert not functional.is_available(session_browser, 'gitweb')
        assert len(functional.find_on_front_page(session_browser,
                                                 'gitweb')) == 0

        functional.logout(session_browser)
        functional.access_url(session_browser, 'gitweb')
        assert functional.is_login_prompt(session_browser)
        assert len(functional.find_on_front_page(session_browser,
                                                 'gitweb')) == 0

    @pytest.mark.backups
    def test_backup_restore(self, session_browser):
        """Test backing up and restoring."""
        _create_repo(session_browser, 'Test-repo', ok_if_exists=True)
        functional.backup_create(session_browser, 'gitweb', 'test_gitweb')
        _delete_repo(session_browser, 'Test-repo')
        functional.backup_restore(session_browser, 'gitweb', 'test_gitweb')
        assert _repo_exists(session_browser, 'Test-repo')
        assert functional.is_available(session_browser, 'gitweb')

    @pytest.mark.parametrize('access', ['public', 'private'])
    @pytest.mark.parametrize('repo_name', ['Test-repo', 'Test-repo.git'])
    @pytest.mark.parametrize('app_status', ['enabled', 'disabled'])
    def test_create_delete_repo(self, session_browser, access, repo_name,
                                app_status):
        """Test creating and deleting a repo and accessing with a git
        client."""
        if app_status == "disabled":
            functional.app_disable(session_browser, 'gitweb')

        _delete_repo(session_browser, repo_name, ignore_missing=True)
        _create_repo(session_browser, repo_name, access)

        assert _repo_exists(session_browser, repo_name, access)

        if app_status == "enabled":
            assert _site_repo_exists(session_browser, repo_name)

            if access == "public":
                assert _repo_is_readable(repo_name)
            else:
                assert not _repo_is_readable(repo_name)

            assert not _repo_is_writable(repo_name)
            assert _repo_is_readable(repo_name, with_auth=True)
            assert _repo_is_writable(repo_name, with_auth=True)

        _delete_repo(session_browser, repo_name)
        assert not _repo_exists(session_browser, repo_name)

    def test_both_private_and_public_repo_exist(self, session_browser):
        """Tests when both private and public repo exist."""
        _create_repo(session_browser, 'Test-repo', 'public', True)
        _create_repo(session_browser, 'Test-repo-private', 'private', True)

        functional.logout(session_browser)
        assert _site_repo_exists(session_browser, 'Test-repo')
        assert not _site_repo_exists(session_browser, 'Test-repo-private')

    @pytest.mark.parametrize('app_status', ['enabled', 'disabled'])
    def test_edit_repo_metadata(self, session_browser, app_status):
        """Test edit repo metadata."""
        if app_status == "disabled":
            functional.app_disable(session_browser, 'gitweb')

        _create_repo(session_browser, 'Test-repo2', 'public',
                     ok_if_exists=True)
        _delete_repo(session_browser, 'Test-repo', ignore_missing=True)
        repo_metadata = {
            'name': 'Test-repo',
            'description': 'Test Description',
            'owner': 'Test Owner',
            'access': 'private',
        }
        _edit_repo_metadata(session_browser, 'Test-repo2', repo_metadata)
        assert _get_repo_metadata(session_browser,
                                  "Test-repo") == repo_metadata

        if app_status == "enabled":
            _create_branch('Test-repo', 'branch1')
            _set_default_branch(session_browser, 'Test-repo', 'branch1')

            assert _get_gitweb_site_default_repo_branch(
                session_browser, 'Test-repo') == 'branch1'


def _create_local_repo(path):
    """Create a local repository."""
    shutil.rmtree(path, ignore_errors=True)
    os.mkdir(path)
    create_repo_commands = [
        'git init -q', 'git config http.sslVerify false',
        'git -c "user.name=Tester" -c "user.email=tester" '
        'commit -q --allow-empty --no-gpg-sign -m "test"'
    ]
    for command in create_repo_commands:
        subprocess.check_call(command, shell=True, cwd=path)


def _create_repo(browser, repo, access=None, ok_if_exists=False):
    """Create repository."""
    if not _repo_exists(browser, repo, access):
        _delete_repo(browser, repo, ignore_missing=True)
        browser.links.find_by_href(
            '/freedombox/apps/gitweb/create/').first.click()
        browser.find_by_id('id_gitweb-name').fill(repo)
        if access == 'private':
            browser.find_by_id('id_gitweb-is_private').check()
        elif access == 'public':
            browser.find_by_id('id_gitweb-is_private').uncheck()
        functional.submit(browser, form_class='form-gitweb')
    elif not ok_if_exists:
        assert False, 'Repo already exists.'


def _create_branch(repo, branch):
    """Add a branch to the repo."""
    repo_url = _get_repo_url(repo, with_auth=True)

    with _gitweb_temp_directory() as temp_directory:
        repo_path = os.path.join(temp_directory, repo)

        _create_local_repo(repo_path)

        add_branch_commands = [['git', 'checkout', '-q', '-b', branch],
                               [
                                   'git', '-c', 'user.name=Tester', '-c',
                                   'user.email=tester', 'commit', '-q',
                                   '--allow-empty', '--no-gpg-sign', '-m',
                                   'test_branch1'
                               ],
                               ['git', 'push', '-q', '-f', repo_url, branch]]
        for command in add_branch_commands:
            subprocess.check_call(command, cwd=repo_path)


def _delete_repo(browser, repo, ignore_missing=False):
    """Delete repository."""
    functional.nav_to_module(browser, 'gitweb')
    if repo.endswith('.git'):
        repo = repo[:-4]
    delete_link = browser.links.find_by_href(
        '/freedombox/apps/gitweb/{}/delete/'.format(repo))
    if delete_link or not ignore_missing:
        delete_link.first.click()
        functional.submit(browser, form_class='form-delete')


def _edit_repo_metadata(browser, repo, metadata):
    """Set repository metadata."""
    functional.nav_to_module(browser, 'gitweb')
    browser.links.find_by_href(
        '/freedombox/apps/gitweb/{}/edit/'.format(repo)).first.click()
    browser.find_by_id('id_gitweb-name').fill(metadata['name'])
    browser.find_by_id('id_gitweb-description').fill(metadata['description'])
    browser.find_by_id('id_gitweb-owner').fill(metadata['owner'])
    if metadata['access'] == 'private':
        browser.find_by_id('id_gitweb-is_private').check()
    else:
        browser.find_by_id('id_gitweb-is_private').uncheck()
    functional.submit(browser, form_class='form-gitweb')


def _get_gitweb_site_default_repo_branch(browser, repo):
    functional.nav_to_module(browser, 'gitweb')
    browser.find_by_css('a[href="/gitweb/{0}.git"]'.format(repo)).first.click()

    return browser.find_by_css('.head').first.text


def _get_repo_metadata(browser, repo):
    """Get repository metadata."""
    functional.nav_to_module(browser, 'gitweb')
    browser.links.find_by_href(
        '/freedombox/apps/gitweb/{}/edit/'.format(repo)).first.click()
    metadata = {}
    for item in ['name', 'description', 'owner']:
        metadata[item] = browser.find_by_id('id_gitweb-' + item).value
    if browser.find_by_id('id_gitweb-is_private').value:
        metadata['access'] = 'private'
    else:
        metadata['access'] = 'public'
    return metadata


def _get_repo_url(repo, with_auth):
    """"Get repository URL"""
    scheme = 'http'
    if _default_url.startswith('https://'):
        scheme = 'https'
    url = _default_url.split(
        '://')[1] if '://' in _default_url else _default_url
    password = 'gitweb_wrong_password'
    if with_auth:
        password = functional.config['DEFAULT']['password']

    return '{0}://{1}:{2}@{3}/gitweb/{4}'.format(
        scheme, functional.config['DEFAULT']['username'], password, url, repo)


def _gitweb_git_command_is_successful(command, cwd):
    """Check if a command runs successfully or gives authentication error"""
    # Tell OS not to translate command return messages
    env = os.environ.copy()
    env['LC_ALL'] = 'C'

    process = subprocess.run(command, capture_output=True, cwd=cwd, env=env,
                             check=False)
    if process.returncode != 0:
        if 'Authentication failed' in process.stderr.decode():
            return False

        process.check_returncode()  # Raise exception
    return True


@contextlib.contextmanager
def _gitweb_temp_directory():
    """Create temporary directory"""
    name = tempfile.mkdtemp(prefix='plinth_test_gitweb_')
    yield name
    shutil.rmtree(name)


def _repo_exists(browser, repo, access=None):
    """Check whether the repository exists."""
    functional.nav_to_module(browser, 'gitweb')
    if not repo.endswith('.git'):
        repo = repo + ".git"
    links_found = browser.links.find_by_href('/gitweb/{}'.format(repo))
    access_matches = True
    if links_found and access:
        parent = links_found.first.find_by_xpath('..').first
        private_icon = parent.find_by_css('.repo-private-icon')
        if access == 'private':
            access_matches = bool(private_icon)
        if access == 'public':
            access_matches = not bool(private_icon)
    return bool(links_found) and access_matches


def _repo_is_readable(repo, with_auth=False):
    """Check if a git repo is readable with git client."""
    url = _get_repo_url(repo, with_auth)
    git_command = ['git', 'clone', '-c', 'http.sslverify=false', url]
    with _gitweb_temp_directory() as cwd:
        return _gitweb_git_command_is_successful(git_command, cwd)


def _repo_is_writable(repo, with_auth=False):
    """Check if a git repo is writable with git client."""
    url = _get_repo_url(repo, with_auth)
    with _gitweb_temp_directory() as temp_directory:
        repo_directory = os.path.join(temp_directory, 'test-project')
        _create_local_repo(repo_directory)
        git_push_command = [
            'git', '-c', 'push.default=current', 'push', '-qf', url
        ]
        return _gitweb_git_command_is_successful(git_push_command,
                                                 repo_directory)


def _set_all_repos_private(browser):
    """Set all repositories private"""
    functional.nav_to_module(browser, 'gitweb')
    public_repos = []
    for element in browser.find_by_css('#gitweb-repo-list .list-group-item'):
        if not element.find_by_css('.repo-private-icon'):
            repo = element.find_by_css('.repo-label').first.text
            public_repos.append(repo)
    for repo in public_repos:
        _set_repo_access(browser, repo, 'private')


def _set_default_branch(browser, repo, branch):
    """Set default branch of the repository."""
    functional.nav_to_module(browser, 'gitweb')
    browser.links.find_by_href(
        '/freedombox/apps/gitweb/{}/edit/'.format(repo)).first.click()
    browser.find_by_id('id_gitweb-default_branch').select(branch)
    functional.submit(browser, form_class='form-gitweb')


def _set_repo_access(browser, repo, access):
    """Set repository as public or private."""
    functional.nav_to_module(browser, 'gitweb')
    browser.links.find_by_href(
        '/freedombox/apps/gitweb/{}/edit/'.format(repo)).first.click()
    if access == 'private':
        browser.find_by_id('id_gitweb-is_private').check()
    else:
        browser.find_by_id('id_gitweb-is_private').uncheck()
    functional.submit(browser, form_class='form-gitweb')


def _site_repo_exists(browser, repo):
    """Check whether the repository exists on Gitweb site."""
    browser.visit('{}/gitweb'.format(_default_url))
    if not repo.endswith('.git'):
        repo = repo + ".git"
    return bool(browser.find_by_css('a[href="/gitweb/{0}"]'.format(repo)))



# SPDX-License-Identifier: AGPL-3.0-or-later
"""Perform distribution upgrade."""

import contextlib
import datetime
import logging
import pathlib
from datetime import timezone
from typing import Generator

import augeas

from plinth import action_utils
from plinth.modules import snapshot as snapshot_module

from . import utils

logger = logging.getLogger(__name__)

OBSOLETE_PACKAGES: list[str] = []

PACKAGES_WITH_PROMPTS = ['firewalld', 'minidlna', 'radicale', 'bind9']

PRE_DEBCONF_SELECTIONS: list[str] = [
    # Tell grub-pc to continue without installing grub again.
    'grub-pc grub-pc/install_devices_empty boolean true'
]

sources_list = pathlib.Path('/etc/apt/sources.list')
temp_sources_list = pathlib.Path('/etc/apt/sources.list.fbx-dist-upgrade')

wait_period_after_release = datetime.timedelta(days=30)

distribution_info: dict = {
    'bullseye': {
        'version': 11,
        'next': 'bookworm',
        'release_date': datetime.datetime(2021, 8, 14, tzinfo=timezone.utc),
    },
    'bookworm': {
        'version': 12,
        'next': 'trixie',
        'release_date': datetime.datetime(2023, 6, 10, tzinfo=timezone.utc),
    },
    'trixie': {
        'version': 13,
        'next': 'forky',
        'release_date': datetime.datetime(2025, 8, 9, tzinfo=timezone.utc),
    },
    'forky': {
        'version': 14,
        'next': 'duke',
        'release_date': None
    },
    'duke': {
        'version': 15,
        'next': None,
        'release_date': None
    },
    'testing': {
        'version': None,
        'next': None,
        'release_date': None
    },
    'unstable': {
        'version': None,
        'next': None,
        'release_date': None
    }
}


def _apt_run(arguments: list[str]):
    """Run an apt command and ensure that output is written to stdout."""
    returncode = action_utils.run_apt_command(arguments,
                                              allow_freedombox_restart=True)
    if returncode:
        raise RuntimeError(
            f'Apt command failed with return code: {returncode}')


def _sources_list_update(old_codename: str, new_codename: str):
    """Change the distribution in /etc/apt/sources.list."""
    logger.info('Upgrading from %s to %s...', old_codename, new_codename)
    aug = augeas.Augeas(flags=augeas.Augeas.NO_LOAD +
                        augeas.Augeas.NO_MODL_AUTOLOAD)
    aug.transform('aptsources', str(sources_list))
    aug.set('/augeas/context', '/files' + str(sources_list))
    aug.set('/augeas/save', 'newfile')  # Save to a new file
    aug.load()

    for match_ in aug.match('*'):
        dist_path = match_ + '/distribution'
        dist = aug.get(dist_path)
        if dist in (old_codename, 'stable'):
            aug.set(dist_path, new_codename)
        elif dist and (dist.startswith(old_codename + '-')
                       or dist.startswith('stable' + '-')):
            new_value = new_codename + '-' + dist.partition('-')[2]
            aug.set(dist_path, new_value)

    aug.save()

    aug_path = sources_list.with_suffix('.list.augnew')
    aug_path.rename(temp_sources_list)


def get_status() -> dict[str, bool | str | None]:
    """Check if a distribution upgrade be performed.

    Check for new stable release, if updates are enabled, and if there is
    enough free space for the dist upgrade.

    Various outcomes:

    - Unattended upgrades are not enabled.
    - Distribution upgrades are not enabled.
    - Not enough free space on the disk to perform dist upgrade.
    - Dist upgrade already running.
    - Codename in base-files package more recent than codename in sources.list.
      Previous run of dist upgrade was interrupted.
    - Could not determine the distribution. Mixed or unknown distribution.
    - On testing/unstable rolling distributions. Nothing to do.
    - On latest stable, no dist upgrade is available. Can upgrade to testing
      (with codename).
    - On old stable, waiting for cool-off period before upgrade. Manual upgrade
      possible.
    - On old stable, ready to do dist upgrade. Manual upgrade possible.

    """
    from plinth.modules import upgrades
    updates_enabled = utils.check_auto()
    dist_upgrade_enabled = upgrades.is_dist_upgrade_enabled()
    has_free_space = utils.is_sufficient_free_space()
    running = action_utils.service_is_running('freedombox-dist-upgrade')

    current_codename = utils.get_sources_list_codename()
    status = {
        'updates_enabled': updates_enabled,
        'dist_upgrade_enabled': dist_upgrade_enabled,
        'has_free_space': has_free_space,
        'running': running,
        'current_codename': current_codename,
        'current_version': None,
        'current_release_date': None,
        'next_codename': None,
        'next_version': None,
        'next_release_date': None,
        'next_action': None,
        'next_action_date': None
    }

    if current_codename in (None, 'testing', 'unstable'):
        return status

    _, base_files_codename = utils.get_current_release()
    if current_codename == 'stable':
        current_codename = base_files_codename

    if current_codename not in distribution_info:
        return status

    current_version = distribution_info[current_codename]['version']
    current_release_date = distribution_info[current_codename]['release_date']
    next_codename = distribution_info[current_codename]['next']
    next_version = None
    next_release_date = None
    if next_codename:
        next_version = distribution_info[next_codename]['version']
        next_release_date = distribution_info[next_codename]['release_date']

    next_action = None
    now = datetime.datetime.now(tz=timezone.utc)
    next_action_date = None
    if next_release_date:
        next_action_date = next_release_date + wait_period_after_release

    if running:
        next_action = None
    elif base_files_codename == next_codename:
        next_action = 'continue'  # Previous run was interrupted
    elif (not next_release_date or not updates_enabled
          or not dist_upgrade_enabled or not has_free_space):
        next_action = None
    elif now >= next_action_date:  # type: ignore
        next_action = 'ready'
    elif now < next_release_date:
        next_action = 'manual'
    else:
        next_action = 'wait_or_manual'

    status.update({
        'current_codename': current_codename,
        'current_version': current_version,
        'current_release_date': current_release_date,
        'next_codename': next_codename,
        'next_version': next_version,
        'next_release_date': next_release_date,
        'next_action': next_action,
        'next_action_date': next_action_date
    })
    return status


@contextlib.contextmanager
def _snapshot_run_and_disable() -> Generator[None, None, None]:
    """Take a snapshot if supported and enabled, then disable snapshots.

    Snapshots shall be re-enabled, if originally enabled, on exiting this
    context manager..
    """
    if not snapshot_module.is_supported():
        logger.info('Snapshots are not supported, skipping taking a snapshot.')
        yield
        return

    reenable = False
    try:
        logger.info('Taking a snapshot before dist upgrade...')
        command = ['snapper', 'create', '--description', 'before dist-upgrade']
        action_utils.run(command, check=True)
        aug = snapshot_module.load_augeas()
        if snapshot_module.is_apt_snapshots_enabled(aug):
            logger.info('Disabling apt snapshots during dist upgrade...')
            action_utils.run([
                '/usr/bin/freedombox-cmd',
                'snapshot',
                'disable_apt_snapshot',
            ], input='{"args": ["yes"], "kwargs": {}}'.encode(), check=True)
            reenable = True
        else:
            logger.info('Apt snapshots already disabled.')

        yield
    finally:
        if reenable:
            logger.info('Re-enabling apt snapshots...')
            action_utils.run([
                '/usr/bin/freedombox-cmd', 'snapshot', 'disable_apt_snapshot'
            ], input='{"args": ["no"], "kwargs": {}}'.encode(), check=True)
        else:
            logger.info('Not re-enabling apt snapshots, as they were disabled '
                        'before dist upgrade.')


@contextlib.contextmanager
def _services_disable():
    """Disable services that are seriously impacted by the upgrade."""
    # If quassel is running during dist upgrade, it may be restarted
    # several times. This causes IRC users to rapidly leave/join
    # channels. Stop quassel for the duration of the dist upgrade.
    logger.info('Stopping quassel service during dist upgrade...')
    with action_utils.service_ensure_stopped('quasselcore'):
        yield
        logger.info('Re-enabling quassel service if previously enabled...')


@contextlib.contextmanager
def _apt_hold_packages():
    """Apt hold some packages during dist upgrade."""
    packages = PACKAGES_WITH_PROMPTS
    packages_string = ', '.join(packages)

    # Hold freedombox package during entire dist upgrade.
    logger.info('Holding freedombox package...')
    with action_utils.apt_hold_freedombox():
        # Hold packages known to have conffile prompts. FreedomBox service
        # will handle their upgrade later.
        logger.info('Holding packages with conffile prompts: %s...',
                    packages_string)
        with action_utils.apt_hold(packages):
            yield
            logger.info(
                'Releasing holds on packages with conffile prompts: %s...',
                packages_string)

        logger.info('Releasing hold on freedombox package...')


def _debconf_set_selections() -> None:
    """Pre-set debconf selections if they are needed for dist upgrade."""
    if PRE_DEBCONF_SELECTIONS:
        logger.info('Setting debconf selections: %s', PRE_DEBCONF_SELECTIONS)
        action_utils.debconf_set_selections(PRE_DEBCONF_SELECTIONS)


def _packages_remove_obsolete() -> None:
    """Remove obsolete packages.

    These may prevent other packages from upgrading.
    """
    if OBSOLETE_PACKAGES:
        logger.info('Removing packages: %s...', OBSOLETE_PACKAGES)
        _apt_run(['remove'] + OBSOLETE_PACKAGES)


def _apt_update():
    """Run 'apt update'."""
    logger.info('Updating Apt cache...')
    _apt_run(['update'])


def _apt_fix():
    """Try to fix any problems with apt/dpkg before the upgrade."""
    logger.info('Fixing any broken apt/dpkg states...')
    action_utils.run(['dpkg', '--configure', '-a'], check=False)
    _apt_run(['--fix-broken', 'install'])


def _apt_autoremove():
    """Run 'apt autoremove'."""
    logger.info('Running apt autoremove...')
    _apt_run(['autoremove'])


def _apt_full_upgrade():
    """Run and check if apt upgrade was successful.

    If the user has changed a configuration file of a package outside of
    FreedomBox, the distribution upgrade process could face a configuration
    file prompt and fail midway. When using unattended-upgrades, these packages
    are not a problem as they would left untouched at an old version and the
    rest of the system would be upgraded. In case of distribution upgrade,
    these packages could cause the distribution upgrade to fail and leave the
    system in an unusable state. Rather than halt distribution upgrade midway
    due to a configuration file prompt, it is better to overwrite with the new
    configuration. Backup copy of the old configuration will be available to
    the user to later merge with the new configuration.

    For packages managed by FreedomBox, packages with configuration file prompt
    will be held back during upgrade and later carefully upgraded with merge.
    These package are not subject to --force-confnew option.
    """
    logger.info('Running apt full-upgrade...')
    _apt_run(['full-upgrade', '-o', 'Dpkg::Options::=--force-confnew'])


def _unattended_upgrades_run():
    """Run unattended-upgrade once more.

    To handle upgrading the freedombox package.
    """
    logger.info('Running unattended-upgrade...')
    action_utils.run(['unattended-upgrade', '--verbose'], check=False)


def _freedombox_restart():
    """Restart FreedomBox service.

    To ensure it is using the latest dependencies.
    """
    logger.info('Restarting FreedomBox service...')
    action_utils.service_restart('plinth')


def _trigger_on_complete():
    """Trigger the on complete step in a separate service."""
    # The dist-upgrade process will be run /etc/apt/sources.list file bind
    # mounted on with a modified file. So, moving modified file to the original
    # file will not be possible. For that, we need to launch a new process with
    # a different systemd service (which does not have the bind mounts).
    logger.info('Triggering on-complete to commit sources.lists')
    action_utils.run([
        'systemd-run', '--unit=freedombox-dist-upgrade-on-complete',
        '--description=Finish up upgrade to new stable Debian release',
        '/usr/bin/freedombox-cmd', 'upgrades', 'dist_upgrade_on_complete',
        '--no-args'
    ], check=True)


def _logging_setup():
    """Log to journal via console logging.

    We need to capture all console logs created by apt and other commands and
    redirect them to journal. This is the default behavior when launching a
    service with systemd-run.

    Avoid double logging to the journal by removing the systemd journal as a
    log handler..
    """
    logging.getLogger(None).removeHandler('journal')


def perform():
    """Perform upgrade to next release of Debian."""
    _logging_setup()
    with _snapshot_run_and_disable(), \
         _services_disable(), \
         _apt_hold_packages():
        _apt_update()
        _apt_fix()
        _debconf_set_selections()
        _packages_remove_obsolete()
        _apt_full_upgrade()
        _apt_autoremove()

    _unattended_upgrades_run()
    _freedombox_restart()
    _trigger_on_complete()


def start_service():
    """Create dist upgrade service and start it."""
    # Cleanup old service
    old_service_path = pathlib.Path(
        '/run/systemd/system/freedombox-dist-upgrade.service')
    if old_service_path.exists():
        old_service_path.unlink(missing_ok=True)
        action_utils.service_daemon_reload()

    status = get_status()
    _sources_list_update(status['current_codename'], status['next_codename'])

    args = [
        '--unit=freedombox-dist-upgrade',
        '--description=Upgrade to new stable Debian release',
        '--property=KillMode=process', '--property=TimeoutSec=72hr',
        f'--property=BindPaths={temp_sources_list}:{sources_list}'
    ]
    action_utils.run(['systemd-run'] + args + [
        'systemd-inhibit', '/usr/bin/freedombox-cmd', 'upgrades',
        'dist_upgrade', '--no-args'
    ], check=True)


def on_complete():
    """Perform cleanup operations."""
    _logging_setup()
    logger.info('Dist upgrade complete.')
    logger.info('Committing changes to /etc/apt/sources.list')
    temp_sources_list.rename(sources_list)



# SPDX-License-Identifier: AGPL-3.0-or-later
"""Views for BIND module."""

from django.contrib import messages
from django.utils.translation import gettext_lazy as _

from plinth.modules import names
from plinth.views import AppView

from . import privileged
from .forms import BindForm


class BindAppView(AppView):  # pylint: disable=too-many-ancestors
    """A specialized view for configuring Bind."""

    app_id = 'bind'
    form_class = BindForm
    template_name = 'bind.html'

    def get_context_data(self, *args, **kwargs):
        """Get/append information for domains and additional names."""
        context = super().get_context_data(**kwargs)

        served_domains = privileged.get_served_domains()
        context['domains_table'] = []
        for key, val in served_domains.items():
            if key == 'localhost.':
                continue

            context['domains_table'].append({
                'type': 'Domain Name',
                'domain_name': key[:-1:],
                'serving': 'Yes',
                'ip_address': ', '.join(val),
            })

        for item in names.components.DomainName.list():
            context['domains_table'].append({
                'type': item.domain_type.display_name,
                'domain_name': item.name,
                'serving': '-',
                'ip_address': ''
            })

        return context

    def get_initial(self):
        """Return the values to fill in the form."""
        initial = super().get_initial()
        initial.update(privileged.get_config())
        return initial

    def form_valid(self, form):
        """Change the configurations of Bind service."""
        data = form.cleaned_data
        old_config = privileged.get_config()

        if old_config['forwarders'] != data['forwarders']:
            privileged.configure(data['forwarders'])
            messages.success(self.request, _('Configuration updated'))

        return super().form_valid(form)



# SPDX-License-Identifier: AGPL-3.0-or-later
"""App component for other apps to use Apache configuration functionality."""

import re
import subprocess

from django.utils.translation import gettext_noop

from plinth import action_utils, app, kvstore
from plinth.diagnostic_check import (DiagnosticCheck,
                                     DiagnosticCheckParameters, Result)
from plinth.privileged import service as service_privileged

from . import privileged


class Webserver(app.LeaderComponent):
    """Component to enable/disable Apache configuration."""

    def __init__(self, component_id: str, web_name: str, kind: str = 'config',
                 urls: list[str] | None = None, expect_redirects: bool = False,
                 last_updated_version: int | None = None):
        """Initialize the web server component.

        component_id should be a unique ID across all components of an app and
        across all components.

        web_name is the primary part of the configuration file path which must
        be enabled/disabled by this component.

        kind is the type of Apache configuration being enabled/disabled. This
        must be 'config' for a configuration in /etc/apache/conf-available/,
        'module' for configuration in /etc/apache2/mods-available/, 'site' for
        configuration in /etc/apache2/sites-available/.

        urls is a list of URLs over which a HTTP services will be available due
        to this component. This list is only used for running diagnostics.

        expect_redirects is a boolean that allows redirects when trying to
        access the URLs during diagnosis of the component.

        last_updated_version is the app version in which the web server
        configuration/site/module file was updated. Using this, web server will
        be automatically reloaded or restarted as necessary during app upgrade.
        """
        super().__init__(component_id)

        self.web_name = web_name
        self.kind = kind
        self.urls = urls or []
        self.expect_redirects = expect_redirects
        self.last_updated_version = last_updated_version or 0

    def is_enabled(self) -> bool:
        """Return whether the Apache configuration is enabled."""
        return action_utils.webserver_is_enabled(self.web_name, kind=self.kind)

    def enable(self) -> None:
        """Enable the Apache configuration."""
        privileged.enable(self.web_name, self.kind)

    def disable(self) -> None:
        """Disable the Apache configuration."""
        privileged.disable(self.web_name, self.kind)

    def diagnose(self) -> list[DiagnosticCheck]:
        """Check if the web path is accessible by clients.

        See :py:meth:`plinth.app.Component.diagnose`.
        """
        results = []
        for url in self.urls:
            if '{host}' in url:
                results.extend(
                    diagnose_url_on_all(url, check_certificate=False,
                                        expect_redirects=self.expect_redirects,
                                        component_id=self.component_id))
            else:
                results.append(
                    diagnose_url(url, check_certificate=False,
                                 component_id=self.component_id))

        return results

    def setup(self, old_version: int):
        """Restart/reload web server if configuration files changed."""
        if not old_version:
            # App is being freshly setup. After setup, app will be enabled
            # which will result in reload/restart of web server.
            return

        if old_version >= self.last_updated_version:
            # Already using the latest configuration. Web server reload/restart
            # is not necessary.
            return

        if not self.app.is_enabled():
            # App is currently disabled, web server will reloaded/restarted
            # when the app is enabled.
            return

        if self.kind == 'module':
            service_privileged.restart('apache2')
        else:
            service_privileged.reload('apache2')


class WebserverRoot(app.FollowerComponent):
    """Component to enable/disable Apache configuration for domain root.

    Each domain has a unique virtual host configuration in Apache. This file
    includes an option configuration file that can dropped in by FreedomBox. If
    an app wants to be hosted on a dedicated domain, it can provide a
    configuration file that is meant to be in the <VirtualHost> section. Using
    this component, the include file fragment for a selected domain can be
    linked to app's configuration file. Then, for the selected domain, the
    app's configuration becomes the domain's root configuration.

    This components uses key/value store to remember the selected domain. When
    the domain changes, the change must be notified using domain_set().
    """

    def __init__(self, component_id: str, web_name: str,
                 expect_redirects: bool = False,
                 last_updated_version: int | None = None):
        """Initialize the web server component for domain root.

        component_id should be a unique ID across all components of an app and
        across all components.

        web_name is the primary part of the configuration file path which must
        be enabled/disabled by this component. The file's path should be
        /etc/apache2/includes/<web_name>.conf.

        expect_redirects is a boolean that allows redirects when trying to
        access the domain URL during diagnosis of the component.

        last_updated_version is the app version in which the web server
        configuration/site/module file was updated. Using this, web server will
        be automatically reloaded or restarted as necessary during app upgrade.
        """
        super().__init__(component_id)

        self.web_name = web_name
        self.expect_redirects = expect_redirects
        self.last_updated_version = last_updated_version or 0

    def enable(self) -> None:
        """Link the Apache site root configuration to app configuration."""
        domain = self.domain_get()
        if domain:
            privileged.link_root(domain, self.web_name)

    def disable(self) -> None:
        """Unlink the Apache site root configuration from app configuration."""
        domain = self.domain_get()
        if domain:
            privileged.unlink_root(domain)

    def _key_get(self) -> str:
        """Return the key used to store the domain in kvstore."""
        return f'{self.component_id}_domain'

    def domain_get(self) -> str | None:
        """Return the currently configured domain name."""
        return kvstore.get_default(self._key_get(), None)

    def domain_set(self, domain: str | None):
        """Set the domain to use with the app."""
        self.disable()
        kvstore.set(self._key_get(), domain)
        if self.app.is_enabled():
            self.enable()

    def diagnose(self) -> list[DiagnosticCheck]:
        """Check if the site root path is accessible by clients.

        See :py:meth:`plinth.app.Component.diagnose`.
        """
        results = []
        domain = self.domain_get()
        if domain:
            results.append(
                diagnose_url(f'https://{domain}', check_certificate=False,
                             component_id=self.component_id))

        return results

    def setup(self, old_version: int):
        """Restart/reload web server if configuration files changed."""
        if not old_version:
            # App is being freshly setup. After setup, app will be enabled
            # which will result in reload/restart of web server.
            return

        if old_version >= self.last_updated_version:
            # Already using the latest configuration. Web server reload/restart
            # is not necessary.
            return

        if not self.app.is_enabled():
            # App is currently disabled, web server will reloaded/restarted
            # when the app is enabled.
            return

        service_privileged.reload('apache2')

    def uninstall(self):
        """Remove the domain configured."""
        kvstore.delete(self._key_get(), ignore_missing=True)


def diagnose_url(url: str, kind: str | None = None,
                 env: dict[str, str] | None = None,
                 check_certificate: bool = True,
                 extra_options: list[str] | None = None,
                 wrapper: str | None = None,
                 expected_output: str | None = None,
                 component_id: str | None = None) -> DiagnosticCheck:
    """Run a diagnostic on whether a URL is accessible.

    Kind can be '4' for IPv4 or '6' for IPv6.
    """
    try:
        return_value = check_url(url, kind, env, check_certificate,
                                 extra_options, wrapper, expected_output)
        result = Result.PASSED if return_value else Result.FAILED
    except FileNotFoundError:
        result = Result.ERROR

    parameters: DiagnosticCheckParameters = {'url': url, 'kind': kind}
    if kind:
        check_id = f'apache-url-kind-{url}-{kind}'
        description = gettext_noop('Access URL {url} on tcp{kind}')
    else:
        check_id = f'apache-url-{url}'
        description = gettext_noop('Access URL {url}')

    return DiagnosticCheck(check_id, description, result, parameters,
                           component_id)


def diagnose_url_on_all(url: str, expect_redirects: bool = False,
                        component_id: str | None = None,
                        **kwargs) -> list[DiagnosticCheck]:
    """Run a diagnostic on whether a URL is accessible."""
    results = []
    for address in action_utils.get_addresses():
        current_url = url.format(host=address['url_address'])
        diagnose_kwargs = dict(kwargs)
        if not expect_redirects:
            diagnose_kwargs.setdefault('kind', address['kind'])

        results.append(
            diagnose_url(current_url, component_id=component_id,
                         **diagnose_kwargs))

    return results


def check_url(url: str, kind: str | None = None,
              env: dict[str, str] | None = None,
              check_certificate: bool = True,
              extra_options: list[str] | None = None,
              wrapper: str | None = None,
              expected_output: str | None = None) -> bool:
    """Check whether a URL is accessible."""
    # When testing a URL with cURL, following any redirections with --location.
    # During those follows, store cookies that have been set and use them for
    # later requests. mod_auth_openidc will set a cookie 'x_csrf' to prevent
    # CSRF attacks and expect this cookie to sent back to it in later requests.
    # If this cookie is not present, it will refuse to perform OIDC Discovery
    # process resulting 404 errors and diagnostic failures for domains that
    # have not been visited by a user recently. --cookie '' means the cURL will
    # use an in-process cookie-jar for storing and retrieving cookies without
    # writing to a file on the disk.
    command = [
        'curl', '--location', '--cookie', '', '--fail', '--write-out',
        '%{response_code}'
    ]

    if kind == '6':
        # extract zone index
        match = re.match(r'(.*://)\[(.*)%(?P<zone>.*)\](.*)', url)
        if match:
            command = command + ['--interface', match.group('zone')]
            url = '{0}[{1}]{2}'.format(*match.group(1, 2, 4))

    command.append(url)

    if wrapper:
        command.insert(0, wrapper)

    if not check_certificate:
        command.append('-k')

    if extra_options:
        command.extend(extra_options)

    if kind:
        command.append({'4': '-4', '6': '-6'}[kind])

    try:
        process = subprocess.run(command, env=env, check=True,
                                 stdout=subprocess.PIPE,
                                 stderr=subprocess.PIPE)
        result = True
        if expected_output and expected_output not in process.stdout.decode():
            result = False
    except subprocess.CalledProcessError as exception:
        result = False
        # Authorization failed is a success
        if exception.stdout.decode().strip() in ('401', '405'):
            result = True

    return result



# SPDX-License-Identifier: AGPL-3.0-or-later
"""FreedomBox app to configure Tor Proxy."""

import json
import logging

from django.urls import reverse_lazy
from django.utils.translation import gettext_lazy as _
from django.utils.translation import gettext_noop

from plinth import app as app_module
from plinth import cfg, frontpage, kvstore, menu
from plinth.daemon import Daemon
from plinth.diagnostic_check import DiagnosticCheck
from plinth.modules.apache.components import diagnose_url
from plinth.modules.backups.components import BackupRestore
from plinth.modules.firewall.components import Firewall
from plinth.modules.users.components import UsersAndGroups
from plinth.package import Packages
from plinth.privileged import service as service_privileged
from plinth.utils import format_lazy

from . import manifest, privileged

logger = logging.getLogger(__name__)

PREINSTALL_CONFIG_KEY = 'torproxy_preinstall_config'

_description = [
    _('Tor is an anonymous communication system. You can learn more '
      'about it from the <a href="https://www.torproject.org/">Tor '
      'Project</a> website. For best protection when web surfing, the '
      'Tor Project recommends that you use the '
      '<a href="https://www.torproject.org/download/download-easy.html.en">'
      'Tor Browser</a>.'),
    format_lazy(
        _('This app provides a web proxy on your {box_name} for internal '
          'networks on TCP port 9050 using the SOCKS protocol. This can be '
          'used by various apps to access the internet via the Tor network. '
          'ISP censorship can be circumvented using upstream bridges.'),
        box_name=_(cfg.box_name))
]


class TorProxyApp(app_module.App):
    """FreedomBox app for Tor Proxy."""

    app_id = 'torproxy'

    _version = 2

    def __init__(self) -> None:
        """Create components for the app."""
        super().__init__()

        info = app_module.Info(app_id=self.app_id, version=self._version,
                               name=_('Tor Proxy'), icon_filename='torproxy',
                               description=_description,
                               manual_page='TorProxy',
                               clients=manifest.clients, tags=manifest.tags,
                               donation_url='https://donate.torproject.org/')
        self.add(info)

        menu_item = menu.Menu('menu-torproxy', info.name, info.icon_filename,
                              info.tags, 'torproxy:index',
                              parent_url_name='apps')
        self.add(menu_item)

        shortcut = frontpage.Shortcut(
            'shortcut-torproxy', info.name, icon=info.icon_filename,
            description=info.description, manual_page=info.manual_page,
            configure_url=reverse_lazy('torproxy:index'), tags=info.tags,
            login_required=True)
        self.add(shortcut)

        packages = Packages('packages-torproxy', [
            'tor', 'tor-geoipdb', 'torsocks', 'obfs4proxy', 'apt-transport-tor'
        ])
        self.add(packages)

        firewall = Firewall('firewall-torproxy-socks', _('Tor Socks Proxy'),
                            ports=['tor-socks'], is_external=False)
        self.add(firewall)

        daemon = Daemon(
            'daemon-torproxy', 'tor@fbxproxy', strict_check=True,
            listen_ports=[(9050, 'tcp4'), (9050, 'tcp6'), (9040, 'tcp4'),
                          (9040, 'tcp6'), (9053, 'udp4'), (9053, 'udp6')])
        self.add(daemon)

        users_and_groups = UsersAndGroups('users-and-groups-torproxy',
                                          reserved_usernames=['debian-tor'])
        self.add(users_and_groups)

        backup_restore = BackupRestore('backup-restore-torproxy',
                                       **manifest.backup)
        self.add(backup_restore)

    def enable(self):
        """Enable app."""
        service_privileged.unmask('tor@fbxproxy')
        super().enable()

    def disable(self):
        """Disable APT use of Tor before disabling."""
        privileged.configure(apt_transport_tor=False)
        super().disable()
        service_privileged.mask('tor@fbxproxy')

    def diagnose(self) -> list[DiagnosticCheck]:
        """Run diagnostics and return the results."""
        results = super().diagnose()
        results.append(_diagnose_url_via_tor('http://www.debian.org', '4'))
        results.append(_diagnose_url_via_tor('http://www.debian.org', '6'))
        results.append(_diagnose_tor_use('https://check.torproject.org', '4'))
        results.append(_diagnose_tor_use('https://check.torproject.org', '6'))
        return results

    def setup(self, old_version):
        """Install and configure the app."""
        super().setup(old_version)
        privileged.setup()
        if not old_version:
            logger.info('Enabling apt-transport-tor')
            config = kvstore.get_default(PREINSTALL_CONFIG_KEY, '{}')
            config = json.loads(config)
            config = {
                'use_upstream_bridges': config.get('use_upstream_bridges'),
                'upstream_bridges': config.get('upstream_bridges'),
                'apt_transport_tor': config.get('apt_transport_tor', True),
            }
            privileged.configure(**config)
            logger.info('Enabling Tor Proxy app')
            self.enable()
            kvstore.delete(PREINSTALL_CONFIG_KEY, ignore_missing=True)

        # Version 2 masks the tor@fbproxy service when disabling the app to
        # prevent the service starting by the Tor master service after system
        # reboot.
        if old_version and old_version < 2 and not self.is_enabled():
            self.disable()

    def uninstall(self):
        """De-configure and uninstall the app."""
        super().uninstall()
        privileged.uninstall()


def _diagnose_url_via_tor(url: str,
                          kind: str | None = None) -> DiagnosticCheck:
    """Diagnose whether a URL is reachable via Tor."""
    result = diagnose_url(url, kind=kind, wrapper='torsocks')
    result.check_id = 'torproxy-url'
    result.description = gettext_noop('Access URL {url} on tcp{kind} via Tor')

    return result


def _diagnose_tor_use(url: str, kind: str | None = None) -> DiagnosticCheck:
    """Diagnose whether webpage at URL reports that we are using Tor."""
    expected_output = 'Congratulations. This browser is configured to use Tor.'
    result = diagnose_url(url, kind=kind, wrapper='torsocks',
                          expected_output=expected_output)
    result.check_id = 'torproxy-using-tor'
    result.description = gettext_noop(
        'Confirm Tor usage at {url} on tcp{kind}')

    return result



# SPDX-License-Identifier: AGPL-3.0-or-later

from django.utils.translation import gettext_lazy as _

from plinth.clients import store_url

clients = [
    {
        'name':
            _('vlc'),
        'platforms': [{
            'type': 'package',
            'os': 'gnu-linux',
            'format': 'deb',
            'name': 'vlc',
        }, {
            'type': 'package',
            'os': 'gnu-linux',
            'format': 'rpm',
            'name': 'vlc',
        }, {
            'type': 'download',
            'os': 'windows',
            'url': 'https://www.videolan.org/vlc/download-windows.html',
        }, {
            'type': 'download',
            'os': 'macos',
            'url': 'https://www.videolan.org/vlc/download-macosx.html',
        }, {
            'type': 'store',
            'os': 'android',
            'store_name': 'google-play',
            'url': store_url('google-play', 'org.videolan.vlc')
        }, {
            'type': 'store',
            'os': 'android',
            'store_name': 'f-droid',
            'url': store_url('f-droid', 'org.videolan.vlc')
        }, {
            'type': 'store',
            'os': 'ios',
            'store_name': 'app-store',
            'url': 'https://apps.apple.com/app/apple-store/id650377962'
        }]
    },
    {
        'name':
            _('kodi'),
        'platforms': [{
            'type': 'package',
            'os': 'gnu-linux',
            'format': 'deb',
            'name': 'kodi',
        }, {
            'type': 'package',
            'os': 'gnu-linux',
            'format': 'rpm',
            'name': 'kodi',
        }, {
            'type': 'download',
            'os': 'windows',
            'url': 'http://kodi.tv/download/',
        }, {
            'type': 'download',
            'os': 'macos',
            'url': 'http://kodi.tv/download/',
        }, {
            'type': 'store',
            'os': 'android',
            'store_name': 'google-play',
            'url': store_url('google-play', 'org.xbmc.kodi')
        }, {
            'type': 'store',
            'os': 'android',
            'store_name': 'f-droid',
            'url': store_url('f-droid', 'org.xbmc.kodi')
        }]
    },
    {
        'name':
            _('yaacc'),
        'platforms': [{
            'type': 'store',
            'os': 'android',
            'store_name': 'f-droid',
            'url': store_url('f-droid', 'de.yaacc')
        }]
    },
    {
        'name':
            _('totem'),
        'platforms': [{
            'type': 'package',
            'os': 'gnu-linux',
            'format': 'deb',
            'name': 'totem',
        }, {
            'type': 'package',
            'os': 'gnu-linux',
            'format': 'rpm',
            'name': 'totem',
        }]
    },
]

# TODO: get all media directories from config file
# for now hard code default media folder.
backup = {
    'data': {
        'files': ['/etc/minidlna.conf'],
        'directories': ['/var/lib/minidlna']
    },
    'services': ['minidlna']
}

tags = [_('Media server'), _('Television'), _('UPnP'), _('DLNA')]



# SPDX-License-Identifier: AGPL-3.0-or-later
"""FreedomBox app to configure matrix-synapse server."""

import logging
import os

from django.urls import reverse_lazy
from django.utils.translation import gettext_lazy as _
from ruamel.yaml.util import load_yaml_guess_indent

from plinth import app as app_module
from plinth import frontpage, menu
from plinth.config import DropinConfigs
from plinth.daemon import Daemon
from plinth.modules.apache.components import Webserver
from plinth.modules.backups.components import BackupRestore
from plinth.modules.coturn.components import TurnConfiguration, TurnConsumer
from plinth.modules.firewall.components import Firewall
from plinth.modules.letsencrypt.components import LetsEncrypt
from plinth.package import Packages, install
from plinth.utils import format_lazy, is_non_empty_file

from . import manifest, privileged

_description = [
    _('<a href="https://matrix.org/docs/guides/faq.html">Matrix</a> is an new '
      'ecosystem for open, federated instant messaging and VoIP. Synapse is a '
      'server implementing the Matrix protocol. It provides chat groups, '
      'audio/video calls, end-to-end encryption, multiple device '
      'synchronization and does not require phone numbers to work. Users on a '
      'given Matrix server can converse with users on all other Matrix '
      'servers via federation.'),
    format_lazy(
        _('Matrix Synapse needs a STUN/TURN server for audio/video calls. '
          'Install the <a href={coturn_url}>Coturn</a> app or configure '
          'an external server.'), coturn_url=reverse_lazy('coturn:index')),
    format_lazy(
        _('<strong>Note:</strong> This app receives frequent feature updates. '
          'It can only be installed if frequent feature updates is enabled in '
          'the <a href="{upgrades_url}">Software Update</a> app.'),
        upgrades_url=reverse_lazy('upgrades:index')),
]

logger = logging.getLogger(__name__)


class MatrixSynapseApp(app_module.App):
    """FreedomBox app for Matrix Synapse."""

    app_id = 'matrixsynapse'

    _version = 11

    def __init__(self) -> None:
        """Create components for the app."""
        super().__init__()

        info = app_module.Info(app_id=self.app_id, version=self._version,
                               depends=['coturn'], name=_('Matrix Synapse'),
                               icon_filename='matrixsynapse',
                               description=_description,
                               manual_page='MatrixSynapse',
                               clients=manifest.clients, tags=manifest.tags)
        self.add(info)

        menu_item = menu.Menu('menu-matrixsynapse', info.name,
                              info.icon_filename, info.tags,
                              'matrixsynapse:index', parent_url_name='apps')
        self.add(menu_item)

        shortcut = frontpage.Shortcut(
            'shortcut-matrixsynapse', info.name, icon=info.icon_filename,
            description=info.description, manual_page=info.manual_page,
            configure_url=reverse_lazy('matrixsynapse:index'),
            clients=info.clients, tags=info.tags, login_required=True)
        self.add(shortcut)

        # Include python3-psycopg2 to prevent accidental uninstall
        # (see issue #2298).
        packages = Packages(
            'packages-matrixsynapse',
            ['matrix-synapse', 'matrix-synapse-ldap3', 'python3-psycopg2'])
        self.add(packages)

        dropin_configs = DropinConfigs('dropin-configs-matrixsynapse', [
            '/etc/apache2/conf-available/matrix-synapse-plinth.conf',
            '/etc/fail2ban/jail.d/matrix-auth-freedombox.conf',
            '/etc/fail2ban/filter.d/matrix-auth-freedombox.conf',
        ])
        self.add(dropin_configs)

        firewall = Firewall('firewall-matrixsynapse', info.name,
                            ports=['matrix-synapse-plinth'], is_external=True)
        self.add(firewall)

        webserver = Webserver('webserver-matrixsynapse',
                              'matrix-synapse-plinth',
                              urls=['https://{host}/_matrix/client/versions'])
        self.add(webserver)

        letsencrypt = LetsEncrypt(
            'letsencrypt-matrixsynapse', domains=get_domains,
            daemons=['matrix-synapse'], should_copy_certificates=True,
            private_key_path='/etc/matrix-synapse/homeserver.tls.key',
            certificate_path='/etc/matrix-synapse/homeserver.tls.crt',
            user_owner='matrix-synapse', group_owner='nogroup',
            managing_app='matrixsynapse')
        self.add(letsencrypt)

        daemon = Daemon('daemon-matrixsynapse', 'matrix-synapse',
                        listen_ports=[(8008, 'tcp4'), (8448, 'tcp4')])
        self.add(daemon)

        backup_restore = BackupRestore('backup-restore-matrixsynapse',
                                       **manifest.backup)
        self.add(backup_restore)

        turn = MatrixSynapseTurnConsumer('turn-matrixsynapse')
        self.add(turn)

    def setup(self, old_version):
        """Install and configure the app."""
        super().setup(old_version)
        if old_version and old_version < 6:
            upgrade()
        else:
            privileged.post_install()

        if old_version and old_version <= 7:
            privileged.fix_public_registrations()

        if not old_version:
            self.enable()

        self.get_component('letsencrypt-matrixsynapse').setup_certificates()

        if not old_version or get_turn_configuration()[1]:
            # Configure STUN/TURN only if there's a valid TLS domain set for
            # Coturn. Do this if app is being freshly installed or if it is
            # previously installed and configured to use STUN/TURN
            # auto-management.
            config = self.get_component(
                'turn-matrixsynapse').get_configuration()
            update_turn_configuration(config, force=True)

    def uninstall(self):
        """De-configure and uninstall the app."""
        super().uninstall()
        privileged.uninstall()


class MatrixSynapseTurnConsumer(TurnConsumer):
    """Component to manage Coturn configuration for Matrix Synapse."""

    def on_config_change(self, config: TurnConfiguration):
        """Add or update STUN/TURN configuration."""
        update_turn_configuration(config)


def upgrade():
    """Upgrade matrix-synapse configuration to avoid conffile prompt."""
    config = privileged.get_config()
    privileged.move_old_conf()
    install(['matrix-synapse'], force_configuration='new', reinstall=True,
            force_missing_configuration=True)
    privileged.post_install()
    privileged.set_config(**config)


def setup_domain(domain_name):
    """Configure a domain name for matrixsynapse."""
    app = app_module.App.get('matrixsynapse')
    app.get_component('letsencrypt-matrixsynapse').setup_certificates(
        [domain_name])
    privileged.setup(domain_name)


def is_setup():
    """Return whether the Matrix Synapse server is setup."""
    return os.path.exists(privileged.SERVER_NAME_PATH)


def get_domains():
    """Return a list of domains this app is interested in."""
    domain = get_configured_domain_name()
    if domain:
        return [domain]

    return []


def get_configured_domain_name():
    """Return the currently configured domain name."""
    if not is_setup():
        return None

    with open(privileged.SERVER_NAME_PATH, encoding='utf-8') as config_file:
        config, _, _ = load_yaml_guess_indent(config_file)

    return config['server_name']


def get_turn_configuration() -> tuple[TurnConfiguration, bool]:
    """Return TurnConfiguration if setup else empty."""
    for file_path, managed in ((privileged.OVERRIDDEN_TURN_CONF_PATH, False),
                               (privileged.TURN_CONF_PATH, True)):
        if is_non_empty_file(file_path):
            with open(file_path, encoding='utf-8') as config_file:
                config, _, _ = load_yaml_guess_indent(config_file)
                return (TurnConfiguration(None, config['turn_uris'],
                                          config['turn_shared_secret']),
                        managed)

    return (TurnConfiguration(), True)


def get_certificate_status():
    """Return the status of certificate for the configured domain."""
    app = app_module.App.get('matrixsynapse')
    status = app.get_component('letsencrypt-matrixsynapse').get_status()
    if not status:
        return 'no-domains'

    return list(status.values())[0]


def update_turn_configuration(config: TurnConfiguration, managed=True,
                              force=False):
    """Update the STUN/TURN server configuration."""
    app = app_module.App.get('matrixsynapse')
    if not force and app.needs_setup():
        return

    privileged.configure_turn(managed, config.to_json())



# SPDX-License-Identifier: AGPL-3.0-or-later
"""
FreedomBox app for Feather Wiki.

This is a FreedomBox-native implementation of a Feather Wiki Nest.
This app doesn't install any Debian packages.
"""

from django.urls import reverse_lazy
from django.utils.translation import gettext_lazy as _

from plinth import app as app_module
from plinth import cfg, frontpage, menu
from plinth.config import DropinConfigs
from plinth.modules.apache.components import Webserver
from plinth.modules.backups.components import BackupRestore
from plinth.modules.firewall.components import Firewall
from plinth.modules.users.components import UsersAndGroups
from plinth.utils import format_lazy

from . import manifest, privileged

_description = [
    format_lazy(
        _('Feather Wiki is a tool to create simple self-contained wikis, each '
          'stored in a single HTML file on your {box_name}. You can use it as '
          'a personal wiki, as a web notebook, or for project documentation.'),
        box_name=_(cfg.box_name)),
    _('Each wiki is a small file. Create as many wikis as you like, such as '
      'one wiki per topic. Customize each wiki to your liking with extensions '
      'and other customization options.'),
    format_lazy(
        _('Feather Wiki is downloaded from {box_name} website and not from '
          'Debian. Wikis need to be upgraded to newer version manually.'),
        box_name=_(cfg.box_name)),
    format_lazy(
        _('Wikis are not public by default, but they can be downloaded for '
          'sharing or publishing. They can be edited by <a href="{users_url}">'
          'any user</a> on {box_name} belonging to the wiki group. '
          'Simultaneous editing is not supported.'), box_name=_(cfg.box_name),
        users_url=reverse_lazy('users:index'))
]


class FeatherWikiApp(app_module.App):
    """FreedomBox app for Feather Wiki."""

    app_id = 'featherwiki'

    _version = 1

    def __init__(self):
        """Create components for the app."""
        super().__init__()

        groups = {'wiki': _('View and edit wiki applications')}

        info = app_module.Info(self.app_id, self._version,
                               name=_('Feather Wiki'),
                               icon_filename='featherwiki',
                               description=_description,
                               manual_page='FeatherWiki',
                               clients=manifest.clients, tags=manifest.tags)
        self.add(info)

        menu_item = menu.Menu('menu-featherwiki', info.name,
                              info.icon_filename, info.tags,
                              'featherwiki:index', parent_url_name='apps')
        self.add(menu_item)

        # The shortcut is a simple directory listing provided by Apache server.
        # Expecting a large number of wiki files, so creating a shortcut for
        # each file (like in ikiwiki's case) will crowd the front page.
        shortcut = frontpage.Shortcut(
            'shortcut-featherwiki', info.name, icon=info.icon_filename,
            description=info.description, manual_page=info.manual_page,
            url='/featherwiki/', clients=info.clients, tags=info.tags,
            login_required=True, allowed_groups=list(groups))
        self.add(shortcut)

        dropin_configs = DropinConfigs('dropin-configs-featherwiki', [
            '/etc/apache2/conf-available/featherwiki-freedombox.conf',
        ])
        self.add(dropin_configs)

        firewall = Firewall('firewall-featherwiki', info.name,
                            ports=['http', 'https'], is_external=True)
        self.add(firewall)

        webserver = Webserver('webserver-featherwiki',
                              'featherwiki-freedombox')
        self.add(webserver)

        users_and_groups = UsersAndGroups('users-and-groups-featherwiki',
                                          groups=groups)
        self.add(users_and_groups)

        backup_restore = BackupRestore('backup-restore-featherwiki',
                                       **manifest.backup)
        self.add(backup_restore)

    def setup(self, old_version=None):
        """Install and configure the app."""
        super().setup(old_version)
        privileged.setup()
        if not old_version:
            self.enable()

    def uninstall(self):
        """Purge directory with all the wikis."""
        super().uninstall()
        privileged.uninstall()


def get_wiki_list():
    """List all the Feather Wiki files."""
    return sorted([
        path.name for path in privileged.wiki_dir.iterdir()
        if path.suffix == '.html'
    ])



# SPDX-License-Identifier: AGPL-3.0-or-later
"""Application manifest for TiddlyWiki."""

from django.utils.translation import gettext_lazy as _

from .privileged import wiki_dir

clients = [{
    'name': _('TiddlyWiki'),
    'platforms': [{
        'type': 'web',
        'url': '/tiddlywiki/'
    }]
}]

backup = {'data': {'directories': [str(wiki_dir)]}}

tags = [
    _('Wiki'),
    _('Note taking'),
    _('Website'),
    _('Journal'),
    _('Digital garden'),
    _('Zettelkasten'),
    _('Quine'),
    _('Non-Debian')
]



# SPDX-License-Identifier: AGPL-3.0-or-later
"""Component to manage a container using podman."""

import contextlib

from django.utils.translation import gettext_noop

from plinth import app, log, privileged
from plinth.daemon import diagnose_port_listening
from plinth.diagnostic_check import (DiagnosticCheck,
                                     DiagnosticCheckParameters, Result)


class Container(app.LeaderComponent, log.LogEmitter):
    """Component to manage a podman container."""

    def __init__(self, component_id: str, name: str, image_name: str,
                 volume_name: str, volume_path: str,
                 volumes: dict[str, str] | None = None,
                 env: dict[str, str] | None = None,
                 binds_to: list[str] | None = None,
                 devices: dict[str, str] | None = None,
                 listen_ports: list[tuple[int, str]] | None = None):
        """Initialize a container component.

        `name` is a string which is the name of the container to create and
        manage. A systemd service unit with the same name is also created.

        `image_name` is a string that represents the repository location from
        which the container images must be pull from.

        `volume_name` is a string with name of the storage volume to create for
        the container to use.

        `volume_path` is a string path on the host machine where the volume
        files for the container is stored.

        `volumes` is a dictionary mapping each string path on the host to a
        string path inside the container. These are bind mounts made available
        inside the container.

        `env` is a dictionary of string key to string values that set the
        environment variables for the processes inside the container to run in.

        `binds_to` is a list of systemd service units that the container's own
        systemd service unit will add BindsTo= and After= dependencies on.

        `devices` is a list of strings with device paths that will be made
        available inside the container. If any of the devices don't exist on
        the host, they will not be added.

        `listen_ports` is a list of tuples containing port number and 'tcp4' or
        'tcp6' network types on which this container is expected to listen on
        after starting the container. This information is used to run
        diagnostic checks on the container.
        """
        super().__init__(component_id)
        self.name = name
        self.image_name = image_name
        self.volume_name = volume_name
        self.volume_path = volume_path
        self.volumes = volumes
        self.env = env
        self.binds_to = binds_to
        self.devices = devices
        self.listen_ports = listen_ports or []

        # For logs
        self.unit = self.name

    def is_enabled(self):
        """Return if the container is enabled."""
        return privileged.container_is_enabled(self.name)

    def enable(self):
        """Run operations to enable and run the container."""
        super().enable()
        privileged.container_enable(self.name)

    def disable(self):
        """Run operations to disable and stop the container."""
        super().disable()
        privileged.container_disable(self.name)

    def is_running(self):
        """Return whether the container service is running."""
        return privileged.is_running(self.name)

    @contextlib.contextmanager
    def ensure_running(self):
        """Ensure a service is running and return to previous state."""
        from plinth.privileged import service as service_privileged
        starting_state = self.is_running()
        if not starting_state:
            service_privileged.enable(self.name)

        try:
            yield starting_state
        finally:
            if not starting_state:
                service_privileged.disable(self.name)

    def setup(self, old_version: int):
        """Bring up and run the container."""
        # Determine whether app should be disabled after setup
        should_disable = old_version and not self.is_enabled()

        privileged.container_setup(self.name, self.image_name,
                                   self.volume_name, self.volume_path,
                                   self.volumes, self.env, self.binds_to,
                                   self.devices)

        if should_disable:
            self.disable()

    def uninstall(self):
        """Remove the container."""
        privileged.container_uninstall(self.name, self.image_name,
                                       self.volume_name, self.volume_path)

    def diagnose(self) -> list[DiagnosticCheck]:
        """Check if the container is running..

        See :py:meth:`plinth.app.Component.diagnose`.
        """
        results = []
        results.append(self._diagnose_unit_is_running())
        for port in self.listen_ports:
            results.append(
                diagnose_port_listening(port[0], port[1], None,
                                        self.component_id))

        return results

    def _diagnose_unit_is_running(self) -> DiagnosticCheck:
        """Check if a daemon is running."""
        check_id = f'container-running-{self.name}'
        result = Result.PASSED if self.is_running() else Result.FAILED

        description = gettext_noop('Container {container_name} is running')
        parameters: DiagnosticCheckParameters = {
            'container_name': str(self.name)
        }

        return DiagnosticCheck(check_id, description, result, parameters,
                               self.component_id)
